
// The MCS locking protocol

Parameter
  n ("number of processes");

Const
  log_np1 = ceil_log(n + 1, 2);

Order
  concat;

Global
  q: range [n + 1];

Clique[n]
  pc:     enum { L0, L1, L2, L3, L4, L5, L6 };
  locked: bool;
  smT:    bool;

Proposition
  Init;
  Bad;

Code

// pc = L0 ==> smT := (q > 0); q := q + 1; pc := L1
UBDD trans0(const StateSpace& S, cushort& p) {
  UBDD trans0 = UBDD::One();

  trans0 &= UBDD::equal(S.pc(p), StateSpace::pc::L0);

  trans0 &= UBDD::equal(S.q(), 0).Ite(UBDD::equal(S._smT(p), false), UBDD::equal(S._smT(p), true));
  trans0 &= UBDD::inc(S.q(), S._q());
  trans0 &= UBDD::equal(S._pc(p), StateSpace::pc::L1);

  trans0 &= UBDD::invariant(S.locked(p));

  return trans0; }

// pc = L1 ==> if smt then pc := L2 else pc := L4
UBDD trans1(const StateSpace& S, cushort& p) {
  UBDD trans1 = UBDD::One();

  trans1 &= UBDD::equal(S.pc(p), StateSpace::pc::L1);

  trans1 &= UBDD::equal(S.smT(p), true).Ite(UBDD::equal(S._pc(p), StateSpace::pc::L2), UBDD::equal(S._pc(p), StateSpace::pc::L4));

  trans1 &= UBDD::invariant(S.q());
  trans1 &= UBDD::invariant(S.locked(p));
  trans1 &= UBDD::invariant(S.smT(p));

  return trans1; }

// pc = L2 ==> locked := true
UBDD trans2(const StateSpace& S, cushort& p) {
  UBDD trans2 = UBDD::One();

  trans2 &= UBDD::equal(S.pc(p), StateSpace::pc::L2);

  trans2 &= UBDD::equal(S._locked(p), true);
  trans2 &= UBDD::equal(S._pc(p), StateSpace::pc::L3);

  trans2 &= UBDD::invariant(S.  q());
  trans2 &= UBDD::invariant(S.smT(p));

  return trans2; }

// pc = L3 ==> repeat until ! locked
UBDD trans3(const StateSpace& S, cushort& p) {
  UBDD trans3 = UBDD::One();

  trans3 &= UBDD::equal(S.    pc(p), StateSpace::pc::L3);
  trans3 &= UBDD::equal(S.locked(p), false);

  trans3 &= UBDD::equal(S._pc(p), StateSpace::pc::L4);

  trans3 &= UBDD::invariant(S.q());
  trans3 &= UBDD::invariant(S.locked(p));
  trans3 &= UBDD::invariant(S.smT(p));

  return trans3; }

// pc = L4 ==> if q = 1 then { q := 0; goto L0 }
UBDD trans4(const StateSpace& S, cushort& p) {
  UBDD trans4 = UBDD::One();

  trans4 &= UBDD::equal(S.pc(p), StateSpace::pc::L4);
  trans4 &= UBDD::equal(S.q(), 1).Ite(UBDD::equal(S._q(), 0) & UBDD::equal(S._pc(p), StateSpace::pc::L0),
				      UBDD::invariant(S.q()) & UBDD::equal(S._pc(p), StateSpace::pc::L5));

  trans4 &= UBDD::invariant(S.locked(p));
  trans4 &= UBDD::invariant(S.smT(p));

  return trans4; }


// pc = L5 & isSpinningInLoop1(j) ==> locked(j) := false
UBDD trans5(const StateSpace& S, cushort& p) {
  UBDD trans5 = UBDD::One();

  trans5 &= UBDD::equal(S.pc(p), StateSpace::pc::L5);

  UBDD someone_spinning_and_freed = UBDD::Zero();
  for (ushort j = 0; j < S.n; j++) {
    UBDD j_spinning_and_freed = UBDD::One();
    j_spinning_and_freed &= UBDD::equal(S.     pc(j), StateSpace::pc::L3);
    j_spinning_and_freed &= UBDD::equal(S. locked(j), true);
    j_spinning_and_freed &= UBDD::equal(S._locked(j), false);
    j_spinning_and_freed &= UBDD::invariant(S.pc(j));
    j_spinning_and_freed &= UBDD::invariant(S.smT(j));
    for (ushort other = 0; other < S.n; other++)  // nobody except p and j changes
      if (other != p && other != j)
	j_spinning_and_freed &= UBDD::invariant(S.pc(other)) & UBDD::invariant(S.locked(other)) & UBDD::invariant(S.smT(other));
    someone_spinning_and_freed |= j_spinning_and_freed; }
  trans5 &= someone_spinning_and_freed;

  trans5 &= UBDD::equal(S._pc(p), StateSpace::pc::L6);

  trans5 &= UBDD::invariant(S.q());
  trans5 &= UBDD::invariant(S.locked(p));
  trans5 &= UBDD::invariant(S.smT(p));

  return trans5; }

// pc = L6 ==> q := q - 1
UBDD trans6(const StateSpace& S, cushort& p) {
  UBDD trans6 = UBDD::One();

  trans6 &= UBDD::equal(S.pc(p), StateSpace::pc::L6);

  trans6 &= UBDD::dec(S.q(), S._q());
  trans6 &= UBDD::equal(S._pc(p), StateSpace::pc::L0);

  trans6 &= UBDD::invariant(S.locked(p));
  trans6 &= UBDD::invariant(S.smT(p));

  return trans6; }

UBDD R(const StateSpace& S) {
  UBDD R = UBDD::Zero();

  for (int p = S.n - 1; p >= 0; p--) {
    UBDD oi = S.invariant_but(p, 0);
    R |= trans0(S, p) & oi;
    R |= trans1(S, p) & oi;
    R |= trans2(S, p) & oi;
    R |= trans3(S, p) & oi;
    R |= trans4(S, p) & oi;
    R |= trans5(S, p); // trans5 changes some other process!
    R |= trans6(S, p) & oi;
  }

  return R; }

// Create initial state: q = 0, local state vars = 0
// For efficiency, go backwards in variable ordering.
UBDD Init(const StateSpace& S) {
  UBDD Init = UBDD::One();
  Init &= UBDD::equal(S.q(), 0);
  for (ushort p = 0; p < S.n; p++) {
    Init &= UBDD::equal(S.pc(p), StateSpace::pc::L0);
    Init &= UBDD::equal(S.locked(p), false);
    Init &= UBDD::equal(S.smT(p), false); }

  return Init; }

// Returns the BDD for the set of bad states, i.e. where at least two processes have pc = L4
UBDD Bad(const StateSpace& S) {
  UBDD Bad = UBDD::Zero();
  for (int p_i = S.n - 1; p_i > 0; p_i--)
    for (int p_j = p_i - 1; p_j >= 0; p_j--)
      Bad |=
	UBDD::equal(S.pc(p_i), StateSpace::pc::L4) &
	UBDD::equal(S.pc(p_j), StateSpace::pc::L4);

  return Bad; }
