%{

// Parsing CTL formulas. "Atomic propositions" are C-style identifiers.

#include <stdlib.h> // for exit

#include <iostream>

#include "functions.hh"
#include "heap.hh"
#include "utool.hh"
#include "model.hh"
#include "model-check.hh"

#define YYSTYPE UBDD*

typedef Heap<UBDD> UBDD_Heap;

UBDD_Heap ubdd_heap;

int yylex  (ModelCheck*, Model&);
int yyerror(ModelCheck*, Model&, char*); // seems yyerror must have same signature as yylex, plus char*

#define YYLEX_PARAM MC,M

%}

%error-verbose
%parse-param {ModelCheck* MC}
%parse-param {Model& M}

// Temporal operators should not have associativity since they are prefix, not infix.

// Binary temporal operators
%nonassoc EU_TOK AU_TOK

// Unary temporal operators
%nonassoc EX_TOK AX_TOK EY_TOK AY_TOK EF_TOK AF_TOK
%nonassoc EG_TOK AG_TOK EP_TOK EX_INF_TOK EY_INF_TOK
%nonassoc FS_TOK BS_TOK INV_FS_TOK INV_BS_TOK

// Binary boolean operators
%left IMPLIES_TOK EQUIV_TOK // => and ==
%left '&' '|'

// Unary boolean operator
%nonassoc '!'

// Constant boolean expressions
%token FALSE_TOK TRUE_TOK

%token IDENTIFIER_TOK STATES_TOK R_TOK SIZE_TOK ALPHA_TOK ORBIT_TOK NEG_TOK

%%

// About reusing BDD* 's: note that if a subformula is an AP, the statement "*$2 = ..." destroys it!
// To prevent this, never assign to *$1, *$2, etc. If a new BDD is created, use "$$ = new UBDD(...)"
// to compute the value of a formula. If not, use "$$ = $2".
// This is the case when converting "(formula)" into "formula", and when converting an AP identifier into its BDD.

spec: formula                  { MC->print(*$1);                                  ubdd_heap.cleanup(); }
  |   INV_FS_TOK formula       { MC->start("INV_FS"); MC->INV_FS(*$2); MC->end(); ubdd_heap.cleanup(); }
  |   INV_BS_TOK formula       { MC->start("INV_BS"); MC->INV_BS(*$2); MC->end(); ubdd_heap.cleanup(); }
  |   R_TOK                    { MC->enumerate_R(); }
  |   SIZE_TOK '(' formula ')' { cout << "number of elements in set computed: ";
	try {
	  unsigned long size = $3->cardinality(UCudd::curr);
	  cout << PPRINT::tabularize(PPRINT::widthify(size)); }
	catch (UBDD::Cardinality_Overflow) {
	  cout << "( > " << PPRINT::tabularize(PPRINT::widthify(UBDD::Cardinality_Overflow::Max)) << " )"; }
	cout << endl;
	ubdd_heap.cleanup(); }

// Binary temporal operators
formula: EU_TOK     formula formula  { MC->start("EU");     $$ = new UBDD(MC->EU(*$2, *$3));   MC->end(*$$); ubdd_heap.insert($$); }
formula: AU_TOK     formula formula  { MC->start("AU");     $$ = new UBDD(MC->AU(*$2, *$3));   MC->end(*$$); ubdd_heap.insert($$); }

// Unary temporal operators
formula: EX_TOK     formula          { MC->start("EX");     $$ = new UBDD(MC->EX(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: AX_TOK     formula          { MC->start("AX");     $$ = new UBDD(MC->AX(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: EY_TOK     formula          { MC->start("EY");     $$ = new UBDD(MC->EY(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: AY_TOK     formula          { MC->start("AY");     $$ = new UBDD(MC->AY(*$2));        MC->end(*$$); ubdd_heap.insert($$); }

formula: EF_TOK     formula          { MC->start("EF");     $$ = new UBDD(MC->EF(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: AF_TOK     formula          { MC->start("AF");     $$ = new UBDD(MC->AF(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: EG_TOK     formula          { MC->start("EG");     $$ = new UBDD(MC->EG(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: AG_TOK     formula          { MC->start("AG");     $$ = new UBDD(MC->AG(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: EP_TOK     formula          { MC->start("EP");     $$ = new UBDD(MC->EP(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: EX_INF_TOK formula          { MC->start("EX_INF"); $$ = new UBDD(MC->EX_INF(*$2));    MC->end(*$$); ubdd_heap.insert($$); }
formula: EY_INF_TOK formula          { MC->start("EY_INF"); $$ = new UBDD(MC->EY_INF(*$2));    MC->end(*$$); ubdd_heap.insert($$); }
formula: FS_TOK     formula          { MC->start("FS");     $$ = new UBDD(MC->FS(*$2));        MC->end(*$$); ubdd_heap.insert($$); }
formula: BS_TOK     formula          { MC->start("BS");     $$ = new UBDD(MC->BS(*$2));        MC->end(*$$); ubdd_heap.insert($$); }

// Unary boolean operator
formula: '!' formula                 { MC->start("!");      $$ = new UBDD(! *$2);              MC->end(*$$); ubdd_heap.insert($$); }

// Binary boolean operators
formula: formula     '|'     formula { MC->start("|");      $$ = new UBDD(*$1     |   *$3) ;   MC->end(*$$); ubdd_heap.insert($$); }
formula: formula     '&'     formula { MC->start("&");      $$ = new UBDD(*$1     &   *$3) ;   MC->end(*$$); ubdd_heap.insert($$); }
formula: formula   EQUIV_TOK formula { MC->start("==");     $$ = new UBDD( $1->  Eqv (*$3));   MC->end(*$$); ubdd_heap.insert($$); }
formula: formula IMPLIES_TOK formula { MC->start("=>");     $$ = new UBDD( $1->IfThen(*$3));   MC->end(*$$); ubdd_heap.insert($$); }

// Functions
formula: NEG_TOK   '(' formula ')'   { MC->start("neg");    $$ = new UBDD(MC->neg  (*$3));     MC->end(*$$); ubdd_heap.insert($$); }
formula: ALPHA_TOK '(' formula ')'   { MC->start("alpha");  $$ = new UBDD(MC->alpha(*$3));     MC->end(*$$); ubdd_heap.insert($$); }
formula: ORBIT_TOK '(' formula ')'   { MC->start("orbit");  $$ = new UBDD(MC->orbit(*$3));     MC->end(*$$); ubdd_heap.insert($$); }

formula: FALSE_TOK                   {                      $$ = new UBDD(UBDD::False());                    ubdd_heap.insert($$); }
formula: TRUE_TOK                    {                      $$ = new UBDD(UBDD::True ());                    ubdd_heap.insert($$); }

formula: IDENTIFIER_TOK  { $$ = $1; } // reuse pointer. Includes "False" and "True"

formula: STATES_TOK      { MC->start("States"); $$ = new UBDD(MC->the_States()); MC->end(*$$); ubdd_heap.insert($$); }

formula: '(' formula ')' { $$ = $2; } // reuse pointer


%%

int yyerror(ModelCheck*, Model&, char*) { std::cerr << "Parse error" << std::endl; exit(1); }
