#include <assert.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <utility>
#include <map>

#include "functions.hh"
#include "triple.hh"

#include "utool.hh"
#include "stateSpace.hh"
#include "manager.hh"
#include "model.hh"
#include "model-check.hh"
#include "model-check-reduced.hh"


extern UCudd* ucudd;


/////////////////////////////////////////////
// Definition of Class ModelCheck_reduced //
/////////////////////////////////////////////

struct PostProcess_Xi: public PostProcess {
  const UBDD& Xi;
  inline PostProcess_Xi(const UBDD& Xi): Xi(Xi) {}
  inline virtual ~PostProcess_Xi() {}
  inline UBDD operator()(const UBDD& Z) { return Xi.existForwIm(Z); } };

struct PostProcess_dynamic: public PostProcess {
  ModelCheck_reduced& MC;
  inline PostProcess_dynamic(ModelCheck_reduced& MC): MC(MC) {}
  inline virtual ~PostProcess_dynamic() {}
  inline UBDD operator()(const UBDD& Z) { return MC.alpha_dynamic(Z); } };

#ifdef __PARTITIONED__

struct PostProcess_partitioned: public PostProcess {
  ModelCheck_reduced& MC;
  inline PostProcess_partitioned(ModelCheck_reduced& MC): MC(MC) {}
  inline virtual ~PostProcess_partitioned() {}
  inline UBDD operator()(const UBDD& Z) { return MC.alpha_partitioned(Z); } };

#endif

ModelCheck_reduced::ModelCheck_reduced(const StateSpace& S, Manager& Man, Model& M, const bool& verbose):
  ModelCheck(S, Man, M, verbose), pp(NULL) {}

void ModelCheck_reduced::set_algorithm(const string& algorithm_s) {
  if      (algorithm_s == "ORBIT")
    algorithm = ORBIT;
  else if (algorithm_s == "MULTIPLE")
    algorithm = MULTIPLE;
  else if (algorithm_s == "DYNAMIC")
    algorithm = DYNAMIC;
#ifdef __PARTITIONED__
  else if (algorithm_s == "PARTITIONED")
    algorithm = PARTITIONED;
#endif
  else
    throw CONTROL::Error("ModelCheck_reduced::set_algorithm: internal: no such algorithm");
  delete_pps();
  switch (algorithm) {
  case ORBIT:       Xi_orbit_();       pp = new PostProcess_Xi         (Xi_orbit);    break;
  case MULTIPLE:    Xi_multiple_();    pp = new PostProcess_Xi         (Xi_multiple); break;
  case DYNAMIC:                        pp = new PostProcess_dynamic    (*this);       break;
#ifdef __PARTITIONED__
  case PARTITIONED: Xi_partitioned_(); pp = new PostProcess_partitioned(*this);       break;
#endif
  default: throw CONTROL::Error("ModelCheck_reduced::set_algorithm: internal: wrong algorithm " + algorithm_s); }}

UBDD ModelCheck_reduced::Rbar_() {
  if (Rbar.getNode() == NULL) {
    Man.start("Rbar");
    Rbar = R;
    Rbar = alpha(Rbar);
    cerr << "finished normalizing left" << endl;
    // Rbar = Alpha(Rbar);
    throw CONTROL::Error("ModelCheck_reduced::Rbar_: Alpha(Rbar) not implemented");
    cerr << "finished normalizing right" << endl;
    Man.end(Rbar); }
  return Rbar; }

UBDD ModelCheck_reduced::Rep_() {
  if (Rep.getNode() == NULL) {
    Rep = UBDD::One();

    if (algorithm == MULTIPLE) {
      Man.start("Rep_multiple");
      // For rotational symmetry, multiple reps seem to make no sense, as there are only n permutations, so we can be exact
      for (int sbn = S.__sbs.size() - 1; sbn >= 0; --sbn)
	switch (S.__sbs[sbn].symmgroup) {
	case Symmblock::FULL:
	  for (ushort p = S.__sbs[sbn].n - 1; p >= 1; --p)
	    Rep &= atMost_as_swapped(0, p, sbn, UCudd::curr);
	  break;
	case Symmblock::RING:
	  for (ushort r = S.__sbs[sbn].n - 1; r >= 1; --r)
	    Rep &= atMost_as_rotated(r, sbn, UCudd::curr);
	  break;
	default: throw CONTROL::Error("ModelCheck_reduced::Rep_: internal"); }}

    else {
      Man.start("Rep_unique");
      for (int sbn = S.__sbs.size() - 1; sbn >= 0; --sbn)
	switch (S.__sbs[sbn].symmgroup) {
	case Symmblock::FULL:
	  // roughly: s <= swap(0, 1) & s <= swap(1, 2) & ... & s <= swap(n-2, n-1)
	  for (int p = S.__sbs[sbn].n - 2; p >= 0; --p)
	    Rep &= atMost_as_swapped(p, p+1, sbn, UCudd::curr);
	  break;
	case Symmblock::RING:
	  // roughly: s <= rotate(1) & s <= rotate(2) & ... & s <= rotate(n-1)
	  for (ushort r = S.__sbs[sbn].n - 1; r >= 1; --r)
	    Rep &= atMost_as_rotated(r, sbn, UCudd::curr);
	  break;
	default: throw CONTROL::Error("ModelCheck_reduced::Rep_: internal"); }}

    Man.end(Rep);
    // Rep.enumerate(Man.print_curr_S);
  }
  return Rep; }

// This definition is based on the theorem that generators for a group acting on a product H x I
// can be obtained as the union of generators for H and generators for I.
UBDD ModelCheck_reduced::Tau_orbit::operator()(const UBDD& Z) {
  UBDD Permute = Z;
  for (int sbn = MC.S.__sbs.size() - 1; sbn >= 0; --sbn)
    switch (MC.S.__sbs[sbn].symmgroup) {
    case Symmblock::FULL: // generators (0,1), (1,2), ...
      for (int p = MC.S.__sbs[sbn].n - 2; p >= 0; --p)
	Permute |= MC.swap(p, p+1, sbn, Z, UCudd::curr);
      break;
    case Symmblock::RING: // generators are the n rotations themselves
      for (ushort r = MC.S.__sbs[sbn].n - 1; r >= 1; --r)
	Permute |= MC.rotate(r, sbn, Z, UCudd::curr);
      break;
    default: throw CONTROL::Error("ModelCheck_reduced::Tau_orbit: internal"); }
  return Permute; }

UBDD ModelCheck_reduced::Theta_orbit_() {
  if (Theta_orbit.getNode() == NULL) {
    Man.start("Theta_orbit");

    // Fixpoint characterization of the orbit relation:
    // Theta(x,y) :: (x=y) \or \exists g: Theta(x,gy)
    // Theta = {(x,X): x and X are equivalent} \subset S x Rep
    Tau_orbit t(*this);
    Theta_orbit = Id_().FP(t, verbose ? "OrbitRel" : "");
    Man.end(Theta_orbit);
    //Theta_orbit.enumerate(Man.print_curr_next_S);
  }
  return Theta_orbit; }

// Unlike the "unique" orbit relation, the "multiple" orbit relation does not require a fixpoint computation.
// We compute the set of pairs (s,t) such that there exists a permutation pi with the following properties:
// (1)  pi(s) = t
// (2a) for all    fully     symmetric blocks, pi has the form swap(0,p) for some p: 0 <= p < n
// (2b) for all rotationally symmetric blocks, pi has the form of a rotation on this block.
// We apply all permutations of the correct form to all symmetry blocks in order, starting with Id.
UBDD ModelCheck_reduced::Theta_multiple_() {
  if (Theta_multiple.getNode() == NULL) {
    Id_();
    Man.start("Theta_multiple");

    Theta_multiple = Id;  // Theta_multiple = {(x,X): x == X | \exists i : i >= 1 : swap{0,i}(x) = X}
    for (int sbn = S.__sbs.size() - 1; sbn >= 0; --sbn) {
      UBDD Z = Theta_multiple;
      switch (S.__sbs[sbn].symmgroup) {
      case Symmblock::FULL:
	for (ushort p = S.__sbs[sbn].n - 1; p >= 1; --p)
	  Theta_multiple |= swap(0, p, sbn, Z, UCudd::curr);
	break;
      case Symmblock::RING: // same as "unique", since same expense
	for (ushort r = S.__sbs[sbn].n - 1; r >= 1; --r)
	  Theta_multiple |= rotate(r, sbn, Z, UCudd::curr);
	break;
      default: throw CONTROL::Error("ModelCheck_reduced::Theta_multiple: internal"); }}
    Man.end(Theta_multiple);
    // Theta_multiple.enumerate(Man.print_curr_next_S);
  }
  return Theta_multiple; }

// Returns Xi_orbit, the representative function: {(x,X): X is the unique representative of x}
UBDD ModelCheck_reduced::Xi_orbit_() {
  if (Xi_orbit.getNode() == NULL) {
    Man.start("Xi_orbit");
    Xi_orbit = Theta_orbit_() & Rep_().shift_to_next();
    Man.end(Xi_orbit);
    cerr << "(clearing Theta_orbit to save memory)" << endl;
    Theta_orbit = UBDD::Zero(); // a good way to free memory??
    // Xi_orbit.enumerate(Man.print_curr_next_S);
  }
  return Xi_orbit; }

// Returns Xi_multiple, a total subset of {(x,X): X is a representative of x}
UBDD ModelCheck_reduced::Xi_multiple_() {
  if (Xi_multiple.getNode() == NULL) {
    Man.start("Xi_multiple");
    Xi_multiple = Theta_multiple_() & Rep_().shift_to_next();
    Man.end(Xi_multiple);
    cerr << "(clearing Theta_multiple to save memory)" << endl;
    Theta_multiple = UBDD::Zero(); // a good way to free memory??
    // Xi_multiple.enumerate(Man.print_curr_next_S);
  }
  return Xi_multiple; }


///////////////////////////////////////////////
//  FUNCTIONS FOR DYNAMIC SYMMETRY REDUCTION //
///////////////////////////////////////////////

UBDD ModelCheck_reduced::one_pass(const UBDD& iterate, const UCudd::Part& part) const {
  UBDD Z = iterate;
  for (int sbn = S.__sbs.size() - 1; sbn >= 0; --sbn)
    switch (S.__sbs[sbn].symmgroup) {
    case Symmblock::FULL:
      for (ushort p = 0; p < S.__sbs[sbn].n - 1; ++p) { // "bottom up" in var ordering didn't seem to be beneficial here
	UBDD Z_bad = Z & greater_than_swapped(p, p+1, sbn, part);
	if (Z_bad != UBDD::Zero()) {
	  UBDD Z_good = Z & ! Z_bad;
	  Z_bad = swap(p, p+1, sbn, Z_bad, part);
	  Z = Z_good | Z_bad; }}
      break;
    case Symmblock::RING:
      for (ushort r = 1; r < S.__sbs[sbn].n; ++r) { // r = rotate amount
	UBDD Z_bad = Z & greater_than_rotated(r, sbn, part);
	if (Z_bad != UBDD::Zero()) {
	  UBDD Z_good = Z & ! Z_bad;
	  Z_bad = rotate(r, sbn, Z_bad, part);
	  Z = Z_good | Z_bad; }}
      break;
    default: throw CONTROL::Error("ModelCheck_reduced::one_pass: internal"); }
  return Z; }

// Swapping (full symmetry)

// Returns {z: z >lex swap_{p,q}(z) }
UBDD ModelCheck_reduced::greater_than_swapped(cushort& p, cushort& q, cushort& sbn, const UCudd::Part& part) const {
  __SAFE_ASSERT__ (p < q);

  const Symmblock& sb = S.__sbs[sbn];
  UBDD greater = UBDD::Zero();
  UBDD same    = UBDD::One();

  // global Idsts
  for (vector<Triple<ushort> >::const_iterator global_idst = sb.global_idsts.begin(); global_idst != sb.global_idsts.end(); ++global_idst) {
    const vector<ushort>& global_idst_bits = S.bits(*global_idst, part);
    greater |= same & UBDD::equal(global_idst_bits, q); // if global = q we have pi(global) = p < q = global
    same    &=     ! (UBDD::equal(global_idst_bits, p) | UBDD::equal(global_idst_bits, q)); }  // for same, g must be neither p nor q

  if (sb.local_idsts[0].empty()) {                        // if no local idsts, z >lex swap{p,q}(z) means l_p > l_q
    const vector<ushort>& p_bits = S.bits(p, sbn, part);
    const vector<ushort>& q_bits = S.bits(q, sbn, part);
    greater |= same & UBDD::greater(p_bits, q_bits); }    // don't need "same" anymore
  else {                                                  // if idsts do exist, we have to go through ALL processes, not just p and q!!
    throw CONTROL::Error("ModelCheck_reduced::greater_than_swapped: internal: not implemented for local idst variables");
//     for (ushort proc = 0; proc < sb.n; ++proc)
//       if (proc == p)
// 	for (ushort local = 0; local < num_of_locals; ++local) {  // step in parallel through corresponding local variables of p and q
// 	  const vector<ushort>& local_p_bits = sb.locals[p][local].get_bits(part);
// 	  const vector<ushort>& local_q_bits = sb.locals[q][local].get_bits(part);
// 	  if (sb.locals[p][local].type == Variable::local) {
// 	    __SAFE_ASSERT__ (sb.locals[q][local].type == Variable::local);
// 	    greater |= same & UBDD::greater(local_p_bits, local_q_bits);
// 	    same    &=        UBDD::equal  (local_p_bits, local_q_bits); }
// 	  else {
// 	    __SAFE_ASSERT__ (sb.locals[q][local].type == Variable::localIdst);
// 	    UBDD ip_is_p = UBDD::equal(local_p_bits, p); UBDD ip_is_q = UBDD::equal(local_p_bits, q); UBDD ip_is_0 = ! (ip_is_p | ip_is_q);
// 	    UBDD iq_is_p = UBDD::equal(local_q_bits, p); UBDD iq_is_q = UBDD::equal(local_q_bits, q); UBDD iq_is_0 = ! (iq_is_p | iq_is_q);
// 	    greater |= same & ip_is_p & iq_is_0 & UBDD::less   (local_q_bits, p);
// 	    greater |= same & ip_is_q & iq_is_q;
// 	    greater |= same & ip_is_q & iq_is_0 & UBDD::less   (local_q_bits, q);
// 	    greater |= same & ip_is_0 & iq_is_p & UBDD::greater(local_p_bits, q);
// 	    greater |= same & ip_is_0 & iq_is_q & UBDD::greater(local_p_bits, p);
// 	    greater |= same & ip_is_0 & iq_is_0 & UBDD::greater(local_p_bits, local_q_bits);
// 	    UBDD new_same = UBDD::Zero();
// 	    new_same |= ip_is_p & iq_is_q;
// 	    new_same |= ip_is_q & iq_is_p;
// 	    new_same |= ip_is_0 & iq_is_0 & UBDD::equal(local_p_bits, local_q_bits);
// 	    same &= new_same; }}
//       else if (proc == q)
// 	continue;            // due to "same", l_p = l_q, i_p = pi(i_q), hence i_q = pi(i_p), and applying pi to q makes no difference
//       else {                 // for processes other than p and q, only idsts. Treatment is as with globalIdsts!
// 	const vector<const Variable*>& localIdsts = sb.localIdsts[proc];
// 	for (vector<const Variable*>::const_iterator local = localIdsts.begin(); local != localIdsts.end(); ++local) {
// 	  __SAFE_ASSERT__ ((*local)->type == Variable::localIdst);
// 	  // cerr << "considering " << (*local)->name << endl;
// 	  const vector<ushort>& local_bits = (*local)->get_bits(part);
// 	  greater |= same & UBDD::equal(local_bits, q);
// 	  same    &= ! (UBDD::equal(local_bits, p) | UBDD::equal(local_bits, q)); }}
  }

  return greater; }

// Returns the BDD for Z(g x, X), where g is the transposition that
// exchanges elements l_i and l_j. I have to exchange:
// x[i][0] with x[j][0], ... , x[i][local_bits - 1] with x[j][local_bits - 1].
//
// In this routine, I resort to "plain Cudd", i.e. without the cuddObj
// extension, for efficiency. cuddObj would require me to pass the BDD sets
// as BDDvectors to the SwapVariables routine, which would dismantle them
// again into DdNode**.
// More on efficiency: I tried on Feb 26, 2004, to inline the code of
// Cudd_bddSwapVariables and to make the xi,xj arrays global, so they have
// to be initialized and freed only once. The results were disappointing:
// Far less than 1% difference in runtime at all, and not even necessarily
// in favor of the streamlined version. So I changed everything back.
UBDD ModelCheck_reduced::swap(cushort& p, cushort& q, cushort& sbn, const UBDD& Z, const UCudd::Part& part) const {
  __SAFE_ASSERT__ (p != q);
  const Symmblock& sb = S.__sbs[sbn];
  cushort& local_bits = sb.local_bits;

  const vector<ushort>& p_bits = S.bits(p, sbn, part);
  const vector<ushort>& q_bits = S.bits(q, sbn, part);

  DdNode** xp = ALLOC(DdNode*, local_bits);
  DdNode** xq = ALLOC(DdNode*, local_bits);

  { ushort i = 0;
    for (vector<ushort>::const_iterator b = p_bits.begin(); b != p_bits.end(); ++b)
      xp[i++] = ucudd->bddVar(*b).getNode(); }

  { ushort i = 0;
    for (vector<ushort>::const_iterator b = q_bits.begin(); b != q_bits.end(); ++b)
      xq[i++] = ucudd->bddVar(*b).getNode(); }

  // the following routine takes the majority of the computation time!
  DdNode* swapped = Cudd_bddSwapVariables(ucudd->getManager(), UBDD(Z).getNode(), xp, xq, local_bits);

  FREE(xp); FREE(xq);
  ucudd->checkReturnValue(swapped);

  UBDD swap = UBDD(ucudd, swapped);

  // for all id sensitive globals in sb:
  for (vector<Triple<ushort> >::const_iterator global_idst = sb.global_idsts.begin(); global_idst != sb.global_idsts.end(); ++global_idst)
    swap = swap_idst(p, q, swap, *global_idst, part);

  // for all id sensitive locals of any process in sb:
  for (ushort proc = 0; proc < sb.n; ++proc)
    for (vector<Triple<ushort> >::const_iterator local_idst = sb.local_idsts[proc].begin(); local_idst != sb.local_idsts[proc].end(); ++local_idst)
      swap = swap_idst(p, q, swap, *local_idst, part);

  return swap; }

// idst is an id-sensitive variable. If its value is p, replace it with q. If its value is q, replace it with p. Otherwise do nothing.
UBDD ModelCheck_reduced::swap_idst(cushort& p, cushort& q, const UBDD& Z, const Triple<ushort>& idst, const UCudd::Part& part) const {
  __SAFE_ASSERT__ (p != q);
  const vector<ushort>& idst_bits = S.bits(idst, part);
  const UBDD&           idst_cube = S.cube(idst, part);
  UBDD idst_is_p = UBDD::equal(idst_bits, p);
  UBDD idst_is_q = UBDD::equal(idst_bits, q);

  UBDD p2q = Z & idst_is_p; // the tuples in Z in which idst has value p
  UBDD q2p = Z & idst_is_q; // the tuples in Z in which idst has value q

  p2q = p2q.ExistAbstract(idst_cube); // "clear" the value of idst
  q2p = q2p.ExistAbstract(idst_cube);

  p2q &= idst_is_q; // "set" the value of idst to q
  q2p &= idst_is_p; // "set" the value of idst to p

  UBDD unchanged = Z & ! (idst_is_p | idst_is_q); // the tuples in Z in which idst has value neither p nor q

  return unchanged | p2q | q2p; }

// Rotating (ring symmetry). We rotate by r to the left: 0 A B C becomes, after 1 rotation, 2 B C A

// Returns { z: z >lex rotate{r}(z) }
UBDD ModelCheck_reduced::greater_than_rotated(cushort& r, cushort& sbn, const UCudd::Part& part) const {
  const Symmblock& sb = S.__sbs[sbn];
  cushort& n = sb.n; // range of idst = 0..(n-1)
  __SAFE_ASSERT__ (r > 0);
  __SAFE_ASSERT__ (r < n);

  // global Idsts. If there are any, the first one determines "greater"; the remaining ones AND the processes are irrelevant!!
  if (! sb.global_idsts.empty()) {
    const vector<ushort>& global_idst_bits = S.bits(sb.global_idsts.front(), part);
    return UBDD::less(global_idst_bits, r); } // left-rotate means: v becomes (v-r) % n: v > (v-r) % n iff v-r < 0 iff v < r

  UBDD greater = UBDD::Zero();
  UBDD same    = UBDD::One();
  if (sb.local_idsts[0].empty())
    for (ushort p = 0; p < n; ++p) {
      cushort p_plus_r = (p+r) % n;
      const vector<ushort>& p_bits        = S.bits(p,        sbn, part);
      const vector<ushort>& p_plus_r_bits = S.bits(p_plus_r, sbn, part);
      greater |= same & UBDD::greater(p_bits, p_plus_r_bits);
      same    &=        UBDD::equal  (p_bits, p_plus_r_bits); }
  else
    throw CONTROL::Error("ModelCheck_reduced::greater_than_rotated: internal: not implemented for local idst variables");

  return greater; }

// Returns Z, with the local states of processes in sbn rotated to the left r positions.
UBDD ModelCheck_reduced::rotate(cushort& r, cushort& sbn, const UBDD& Z, const UCudd::Part& part) const {
  const Symmblock& sb = S.__sbs[sbn];
  cushort& local_bits = sb.local_bits;
  cushort& n          = sb.n; // range of idst = 0..(n-1)
  __SAFE_ASSERT__ (r > 0);
  __SAFE_ASSERT__ (r < n);

  int* permute = new int[ucudd->ReadSize()];
  // Start by claiming all variables remain the same
  for (ushort b = 0; b < ucudd->ReadSize(); ++b)
    permute[b] = b;
  for (ushort p = 0; p < n; ++p) {
    cushort p_plus_r = (p+r) % n;
    for (ushort b = 0; b < local_bits; ++b) {
      cushort i = S.bit(p,        sbn, b, part);
      cushort j = S.bit(p_plus_r, sbn, b, part);
      permute[j] = i; }}

  UBDD rotate = Z.Permute(permute);

  delete permute;

  // for all id sensitive globals in sb:
  for (vector<Triple<ushort> >::const_iterator global_idst = sb.global_idsts.begin(); global_idst != sb.global_idsts.end(); ++global_idst)
    rotate = rotate_idst(r, sbn, rotate, *global_idst, part);

  // for all id sensitive locals of any process in sb:
  for (ushort proc = 0; proc < sb.n; ++proc)
    for (vector<Triple<ushort> >::const_iterator local_idst = sb.local_idsts[proc].begin(); local_idst != sb.local_idsts[proc].end(); ++local_idst)
      rotate = rotate_idst(r, sbn, rotate, *local_idst, part);

  return rotate; }

// idst is an id-sensitive variable. Rotate to the left: subtract r from idst's value
UBDD ModelCheck_reduced::rotate_idst(cushort& r, cushort& sbn, const UBDD& Z, const Triple<ushort>& idst, const UCudd::Part& part) const {
  cushort& n = S.__sbs[sbn].n; // range of idst = 0..(n-1)
  __SAFE_ASSERT__ (r > 0);
  __SAFE_ASSERT__ (r < n);
  const vector<ushort>& idst_bits = S.bits(idst, part);
  const UBDD&           idst_cube = S.cube(idst, part);

  UBDD rotate_idst = UBDD::Zero();
  for (ushort p = 0; p < n; ++p) {                       // possible values of idst
    cushort p_minus_r = ((n+p)-r) % n;                   // CAREFUL WITH - AND %: (2-5) % 4 is -3! So make it positive first
    UBDD idst_is_p         = UBDD::equal(idst_bits, p);
    UBDD idst_is_p_minus_r = UBDD::equal(idst_bits, p_minus_r);
    UBDD p2p_minus_r = Z & idst_is_p;                    // the tuples in Z in which idst has value p
    p2p_minus_r  = p2p_minus_r.ExistAbstract(idst_cube); // "clear" the value of idst
    p2p_minus_r &= idst_is_p_minus_r;                    // "set"   the value of idst to p-r
    rotate_idst |= p2p_minus_r; }
  return rotate_idst; }

#ifdef __PARTITIONED__

class BitVector {
private:
  Vector<bool> v;
  cushort n; // v.size()
   ushort l; // index of last 1 that is followed by 0, if exists; >=n otherwise
   ushort a; // number of 1s following v[l], if l < n
public:
  void reset(cushort& k);
  inline BitVector(cushort& n, cushort& k = 0): v(n), n(n) { if (k > n) throw CONTROL::Error("k > n"); reset(k); }
  bool operator++();
  inline const vector<bool>& the_vector() const { return v; } };

void BitVector::reset(cushort& k) {
  for (ushort i = 0; i < k; ++i)
    v[i] = true;
  for (ushort i = k; i < n; ++i)
    v[i] = false;
  l = ( k == n ? n : k-1 ); // note: if k == 0, l will become infty and hence >= n, disallowing further increments
  a = 0; }

bool BitVector::operator++() {
  if (l >= n)
    return false;
  v[ l ] = false;
  v[l+1] = true;
  if (a < n - l - 2) { // 0 exists at or after position l+2. We don't need to move left to find the new l
    ushort ap = a; // don't detroy a, we need it below
    for (ushort i = l + 2; i < n; ++i)
      if (ap > 0) {
	v[i] = true;
	--ap; }
      else
	v[i] = false;
    l = l + 1 + a;
    a = 0; }
  else {
    ++a; --l;
    while ( l >= n ? false : !v[l] )
      --l; }
  return true; }

void ModelCheck_reduced::Xi_partitioned_() {
  Man.start("Xi_partitioned");
  cushort& n = S.__sbs[0].n;
  BitVector V(n);
  for (ushort p = 0; p < n; ++p) {
    cerr << "p: " << p << endl;
    for (ushort L = 0; L < S.l; ++L) {
      // cerr << "L: " << L << endl;

      UBDD atMost = UBDD::Zero();
      V.reset(p + 1);
      do {
	UBDD for_V = UBDD::One();
	for (ushort q = 0; q < n; ++q)
	  if (V.the_vector()[q])
	    for_V &= UBDD::atMost(S.L(q), L);
	atMost |= for_V; }
      while (V++);

      UBDD atLeast = UBDD::Zero();
      V.reset(n - p);
      do {
	UBDD for_V = UBDD::One();
	for (ushort q = 0; q < n; ++q)
	  if (V.the_vector()[q])
	    for_V &= UBDD::atLeast(S.L(q), L);
	atLeast |= for_V; }
      while (V++);

      UBDD result = atMost & atLeast;
      // result.enumerate(Man.print_curr_next_S);

      // store result into cache
      UU_pair pL = std::make_pair(p, L);
      partition_cache.insert(std::make_pair(pL, result)); }}
  __SAFE_ASSERT__ (partition_cache.size() == Partition_Cache::size_type(n * S.l));
  Man.end(); }

UBDD ModelCheck_reduced::alpha_partitioned(const UBDD& Z) {
  UBDD result = Z;
  for (ushort b = 0; b < S.bit(0, 0, 0, UCudd::curr); ++b) // global variables
    result &= ucudd->bddVar(b).Eqv(ucudd->bddVar(b+1));
  for (int p = S.__sbs[0].n - 1; p >= 0; --p) {
    UBDD for_p = UBDD::Zero();
    for (int L = S.l - 1; L >= 0; --L) {
      UU_pair pL = std::make_pair(p, L);
      Partition_Cache::const_iterator found = partition_cache.find(pL);
      for_p |= found->second & UBDD::equal(S._L(p), L); }
    result &= for_p; }
  return result.existsCurrState().shift_to_curr(); }

#endif

// performs forward search
void ModelCheck_reduced::INV_FS(const UBDD& Good) {
  List<UBDD> images;
  UBDD Init = M.AP("Init"), Bad = !Good, Unexplored = Init, Reached = Unexplored; // '!' instead of neg. This is cheating, but correct here
  if (! test_and_update_FS(Unexplored, Bad, images)) {
    counterExample_FS(images);
    return; }
  Report r("INV_FS", verbose);
  while (! Unexplored.empty()) {
    UBDD Successors = EY(Unexplored);
    Successors = (*pp)(Successors);
    Unexplored = Successors & (! Reached); // we can use ! here instead of neg since Successors \subset Rep
    if (! test_and_update_FS(Unexplored, Bad, images)) {
      counterExample_FS(images);
      return; }
    Reached |= Unexplored;
    r.print(Unexplored, Reached); }

  cout << "  -- INVARIANT HOLDS" << endl; }

// the last UBDD in the images list is the set of reached Bad states
void ModelCheck_reduced::counterExample_FS(const List<UBDD>& images) const {
  cout << "  *** INVARIANT FAILED; generating counter example trace of edge length " << images.size() - 1 << " ... "; cout.flush();
  __SAFE_ASSERT__ (! images.empty());
  UBDD error = images.back().randomCube(0, UCudd::curr);
  list<UBDD> path;
  path.push_front(error);
  for (List<UBDD>::const_reverse_iterator image = images.succ(images.rbegin()); image != images.rend(); ++image) { // note: starts at 2nd element
    UBDD EX_error = EX(error);
    UBDD error_rep = (alpha(EX_error) & (*image)).randomCube(0, UCudd::curr);
    error = (orbit(error_rep) & EX_error).randomCube(0, UCudd::curr);
    path.push_front(error); }
  print(path); }

// performs backward search
void ModelCheck_reduced::INV_BS(const UBDD& Good) {
  List<UBDD> images;
  UBDD Init = M.AP("Init"), Bad = neg(Good), Unexplored = Bad, Reached = Unexplored; // (DO NOT replace neg by ! here)
  if (! test_and_update_BS(Unexplored, Init, images)) {
    counterExample_BS(images);
    return; }
  Report r("INV_BS", verbose);
  while (! Unexplored.empty()) {
    UBDD Successors = EX(Unexplored);
    Successors = (*pp)(Successors);
    Unexplored = Successors & (! Reached); // see comment in INV_FS
    if (! test_and_update_BS(Unexplored, Init, images)) {
      counterExample_BS(images);
      return; }
    Reached |= Unexplored;
    r.print(Unexplored, Reached); }

  cout << "  -- INVARIANT HOLDS" << endl; }

// the first UBDD in the images list is the set of reached Init states
void ModelCheck_reduced::counterExample_BS(const List<UBDD>& images) const {
  cout << "  *** INVARIANT FAILED; generating counter example trace of edge length " << images.size() - 1 << " ... "; cout.flush();
  __SAFE_ASSERT__ (! images.empty());
  UBDD error = images.front().randomCube(0, UCudd::curr);
  list<UBDD> path;
  path.push_back(error);
  for (list<UBDD>::const_iterator image = images.succ(images.begin()); image != images.end(); ++image) { // note: starts at 2nd element
    UBDD EY_error = EY(error);
    UBDD error_rep = (alpha(EY_error) & (*image)).randomCube(0, UCudd::curr);
    error = (orbit(error_rep) & EY_error).randomCube(0, UCudd::curr);
    path.push_back(error); }
  print(path); }

void ModelCheck_reduced::test(const UBDD& Z) {}
