#ifndef __MODELCHECK__SPECIFC__
#define __MODELCHECK__SPECIFC__

#include <string>
#include <vector>

#include "functions.hh"
#include "triple.hh"

#include "utool.hh"
#include "stateSpace.hh"
#include "manager.hh"
#include "model.hh"
#include "model-check.hh"


struct PostProcess { virtual UBDD operator()(const UBDD&) = 0; virtual inline ~PostProcess() {}; };

class ModelCheck_reduced: public ModelCheck {
public:
  typedef enum { ORBIT, MULTIPLE, DYNAMIC } Algorithm; // PARTITIONED
private:
  Algorithm algorithm;

  PostProcess *pp; // these must be pointers to enable dynamic binding of PostProcess::operator()

#ifdef __PARTITIONED__
  typedef std::pair<ushort,ushort> UU_pair;
  typedef map<UU_pair,UBDD> Partition_Cache;
#endif

  UBDD Id;              // for all BDD variables, curr=next. Needed in fixpoint routine to build orbit relation
  UBDD Rbar;            // normally not used (experimental)
  UBDD Rep;
  UBDD Theta_orbit;
  UBDD Theta_multiple;
  UBDD Xi_orbit;
  UBDD Xi_multiple;
#ifdef __PARTITIONED__
  Partition_Cache partition_cache;
#endif

  UBDD Rbar_();
  UBDD Rep_();
  UBDD Xi_orbit_();
  UBDD Xi_multiple_();
#ifdef __PARTITIONED__
  void Xi_partitioned_();
#endif

public:
  ModelCheck_reduced(const StateSpace& S, Manager& Man, Model& M, const bool& verbose);
  void set_algorithm(const string& algorithm_s);
  inline void delete_pps() { if (pp != NULL) delete pp; }
  inline ~ModelCheck_reduced() { delete_pps(); }

  void test(const UBDD& Z);

private:
  struct Tau_orbit: public Tau {
    const ModelCheck_reduced& MC;
    Tau_orbit(const ModelCheck_reduced& MC): MC(MC) {}
    UBDD operator()(const UBDD& Z); };

  inline UBDD Id_() { if (Id.getNode() == NULL) Id = UBDD::identity(); return Id; }

  UBDD Theta_orbit_();
  UBDD Theta_multiple_();

  inline UBDD neg  (const UBDD& Z)       { return (!Z) & Rep_(); } // negation in the quotient
  inline UBDD alpha(const UBDD& Z) const { return (*pp)(Z); }
  inline UBDD orbit(const UBDD& Z) const { Tau_orbit t(*this); return Z.FP(t, "orbit"); } // {s': \exists s \in S: s' == s}, a superset of S

  /////////////////////// Routines for Dynamic Symmetry Reduction ///////////////////////////

  inline UBDD alpha_dynamic(const UBDD& Z) const;
  UBDD one_pass(const UBDD& iterate, const UCudd::Part& part) const;

         UBDD swap                (cushort& p, cushort& q, cushort& sbn, const UBDD& Z,                             const UCudd::Part& part) const;
         UBDD swap_idst           (cushort& p, cushort& q,               const UBDD& Z, const Triple<ushort>& idst, const UCudd::Part& part) const;
         UBDD greater_than_swapped(cushort& p, cushort& q, cushort& sbn,                                            const UCudd::Part& part) const;
  inline UBDD atMost_as_swapped   (cushort& p, cushort& q, cushort& sbn,                                            const UCudd::Part& part) const {
    return ! greater_than_swapped(p, q, sbn, part); }

         UBDD rotate              (cushort& r, cushort& sbn, const UBDD& Z,                             const UCudd::Part& part) const;
         UBDD rotate_idst         (cushort& r, cushort& sbn, const UBDD& Z, const Triple<ushort>& idst, const UCudd::Part& part) const; // we need the sbn here, unlike above
         UBDD greater_than_rotated(cushort& r, cushort& sbn,                                            const UCudd::Part& part) const;
  inline UBDD atMost_as_rotated   (cushort& r, cushort& sbn,                                            const UCudd::Part& part) const {
    return ! greater_than_rotated(r, sbn, part); }

  /////////////////////// Routines for Partitioned Symmetry Reduction ///////////////////////////

#ifdef __PARTITIONED__
  UBDD alpha_partitioned(const UBDD& Z);
#endif

public:

  UBDD EF    (const UBDD& f);
  UBDD EG    (const UBDD& f);
  UBDD EU    (const UBDD& f, const UBDD& g);
  UBDD EP    (const UBDD& f);
  UBDD AF    (const UBDD& f);
  UBDD AG    (const UBDD& f);
  UBDD AU    (const UBDD& f, const UBDD& g);
  UBDD EX_INF(const UBDD& f);
  UBDD EY_INF(const UBDD& f);
  UBDD FS    (const UBDD& f);
  UBDD BS    (const UBDD& f);

  void INV_FS(const UBDD& Good);
  void INV_BS(const UBDD& Good);
  void counterExample_FS(const List<UBDD>& images) const;
  void counterExample_BS(const List<UBDD>& images) const;

  friend struct AX_Image_reduced;
  friend struct Tau_alpha_dynamic;
  friend struct PostProcess_dynamic;
#ifdef __PARTITIONED__
  friend struct PostProcess_partitioned;
#endif
};


//////////////////////////////////////////////
// Declaration of "reduced" Image operators //
//////////////////////////////////////////////

struct EX_Image_reduced: public EX_Image {
  PostProcess& pp;
  inline EX_Image_reduced(const UBDD& R, PostProcess& pp): EX_Image(R), pp(pp) {}
  inline UBDD operator()(const UBDD& Z) { UBDD EXZ = R.EX(Z); return pp(EXZ); } };

struct AX_Image_reduced: public AX_Image {
  PostProcess& pp;
  ModelCheck_reduced& MC;
  inline AX_Image_reduced(const UBDD& R, PostProcess& pp, ModelCheck_reduced& MC): AX_Image(R), pp(pp), MC(MC) {}
  inline UBDD operator()(const UBDD& Z) { UBDD not_Z = MC.neg(Z); UBDD EXZ = R.EX(not_Z); pp(EXZ); return MC.neg(EXZ); } };

struct EY_Image_reduced: public EY_Image {
  PostProcess& pp;
  inline EY_Image_reduced(const UBDD& R, PostProcess& pp): EY_Image(R), pp(pp) {}
  inline UBDD operator()(const UBDD& Z) { UBDD EYZ = R.EY(Z); return pp(EYZ); } };

inline UBDD ModelCheck_reduced::EF    (const UBDD& f)                { EX_Image_reduced I(R, *pp);        return UBDD::EF    (f,    I, verbose); }
inline UBDD ModelCheck_reduced::EG    (const UBDD& f)                { EX_Image_reduced I(R, *pp);        return UBDD::EG    (f,    I, verbose); }
inline UBDD ModelCheck_reduced::EU    (const UBDD& f, const UBDD& g) { EX_Image_reduced I(R, *pp);        return UBDD::EU    (f, g, I, verbose); }
inline UBDD ModelCheck_reduced::EP    (const UBDD& f)                { EY_Image_reduced I(R, *pp);        return UBDD::EP    (f,    I, verbose); }
inline UBDD ModelCheck_reduced::AF    (const UBDD& f)                { AX_Image_reduced I(R, *pp, *this); return UBDD::AF    (f,    I, verbose); }
inline UBDD ModelCheck_reduced::AG    (const UBDD& f)                { AX_Image_reduced I(R, *pp, *this); return UBDD::AG    (f,    I, verbose); }
inline UBDD ModelCheck_reduced::AU    (const UBDD& f, const UBDD& g) { AX_Image_reduced I(R, *pp, *this); return UBDD::AU    (f, g, I, verbose); }
inline UBDD ModelCheck_reduced::EX_INF(const UBDD& f)                { EX_Image_reduced I(R, *pp);        return UBDD::EX_INF(f,    I, verbose); }
inline UBDD ModelCheck_reduced::EY_INF(const UBDD& f)                { EY_Image_reduced I(R, *pp);        return UBDD::EY_INF(f,    I, verbose); }
inline UBDD ModelCheck_reduced::FS    (const UBDD& f)                { EY_Image_reduced I(R, *pp);        return UBDD::FS    (f,    I, verbose); }
inline UBDD ModelCheck_reduced::BS    (const UBDD& f)                { EX_Image_reduced I(R, *pp);        return UBDD::BS    (f,    I, verbose); }

struct Tau_alpha_dynamic: public Tau {
  const ModelCheck_reduced& MC;
  inline Tau_alpha_dynamic(const ModelCheck_reduced& MC): MC(MC) {}
  inline UBDD operator()(const UBDD& Z) { return MC.one_pass(Z, UCudd::curr); } };

inline UBDD ModelCheck_reduced::alpha_dynamic(const UBDD& Z) const { Tau_alpha_dynamic t(*this); return Z.FP(t, "alpha_dynamic"); } // always non-verbose

#endif
