#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include <algorithm>
#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <set>
#include <list>

#include "parse.hh"
#include "functions.hh"
#include "symbint.hh"

#include "sviss-symtab.hh"

const Set<string> Parse::RESERVED("invariant, invariant_but, invariant_sbn, invariant_all, States", ", ");

Parse::Parse(const string& ifilename, const string& verbose): ifilename(ifilename), verbose(verbose == "true"), line_number(1), global_varCount(0), sb_cnt(-1), APs(" "),
							      s_heap("string"), i_heap("SymbInt"), l_heap("list<string>"),
							      parameter_decls(""), parameter_inits(""), parameter_passes("") {
  open_verbose(ifile, ifilename);

  open_verbose(run_parameter_cpp,             "run-parameter.cpp");
  open_verbose(run_construct_stateSpace_cpp,  "run-construct-stateSpace.cpp");
  open_verbose(stateSpace_user_hpp,           "stateSpace-user.hpp");
  open_verbose(stateSpace_user_cc,            "stateSpace-user.cc");
  open_verbose(R_APs_hpp,                     "R-APs.hpp");
  open_verbose(R_APs_cc,                      "R-APs.cc"); }

Parse::~Parse() {
  close_verbose(R_APs_cc,                     "R-APs.cc");
  close_verbose(R_APs_hpp,                    "R-APs.hpp");
  close_verbose(stateSpace_user_cc,           "stateSpace-user.cc");
  close_verbose(stateSpace_user_hpp,          "stateSpace-user.hpp");
  close_verbose(run_construct_stateSpace_cpp, "run-construct-stateSpace.cpp");
  close_verbose(run_parameter_cpp,            "run-parameter.cpp");

  close_verbose(ifile, ifilename);

  l_heap.cleanup( verbose ? L_Heap::VERBOSE : L_Heap::SILENT );
  i_heap.cleanup( verbose ? I_Heap::VERBOSE : I_Heap::SILENT );
  s_heap.cleanup( verbose ? S_Heap::VERBOSE : S_Heap::SILENT ); }

void Parse::remove_verbose() {
  remove_verbose(R_APs_cc,                     "R-APs.cc");
  remove_verbose(R_APs_hpp,                    "R-APs.hpp");
  remove_verbose(stateSpace_user_cc,           "stateSpace-user.cc");
  remove_verbose(stateSpace_user_hpp,          "stateSpace-user.hpp");
  remove_verbose(run_construct_stateSpace_cpp, "run-construct-stateSpace.cpp");
  remove_verbose(run_parameter_cpp,            "run-parameter.cpp");
  close_verbose(ifile, ifilename); }

void Parse::print_symtab(ostream& out) const {
  out << "Contents of Symbol Table:" << endl;
  sviss_symtab.print(2, out);
  out << "Number of global variables: " << global_varCount << endl
      << "Symmetry block information:" << endl;
  for (ushort sbn = 0; sbn < sb_ns.size(); ++sbn)
    out << "  Symmetry block " << sbn << " with " << ( sg_ids[sbn] == CLIQUE ? "full" : "ring" ) << " symmetry group, " << sb_ns[sbn] << " processes and "
	<< sb_varCounts[sbn] << " local variable" << ( sb_varCounts[sbn] == 1 ? "" : "s" ) << " each" << endl;
  out << "Atomic propositions:" << endl
      << "  " << APs
      << endl; }

void Parse::parameter_decl(const string& symname, const string& meaning, const string& default_value) {
  sviss_symtab.insert_parameter(symname, new Parameter(meaning, default_value));
  stateSpace_user_hpp << "  cushort " << symname << ";" << endl;
  run_parameter_cpp << "  I.add_argument(\"" << symname << "\"" << ", " << meaning << ", \"\", " << default_value << ");" << endl;
  const string separator = ( parameter_decls.empty() ? "" : ", " ); // same separator in all cases!
  parameter_decls  += separator + "cushort& " + symname;
  parameter_inits  += separator + symname + "(" + symname + ")";
  parameter_passes += separator + "atoi(I.arg_value(\"" + symname + "\").c_str())"; }

void Parse::constant_decl(const string& symname, const SymbInt& value) {
  sviss_symtab.insert_constant(symname, new Constant(value)); // this enters the *expr* used to compute the constant into the symtab!
  stateSpace_user_hpp << "  cushort " << symname << ";" << endl;
  const string separator = ( constant_inits.empty() ? "" : ", " );
  constant_inits += separator + symname + "(" + string(value) + ")"; }

void Parse::stateSpace() {
  const string separator = ( parameter_inits.empty() || constant_inits.empty() ? "" : ", " );
  const string inits     = parameter_inits + separator + constant_inits;
  const string colon     = ( inits.empty() ? "" : ": " );

  stateSpace_user_hpp << endl
		      << "  StateSpace(" << parameter_decls << ")" << colon << inits << " { __initialize(); __make_States(); }" << endl;

  run_construct_stateSpace_cpp << "  StateSpace S" << ( parameter_passes.empty() ? "" : "(" + parameter_passes + ")" ) << ";" << endl; // no () if empty!

  const bool concat = ( sviss_symtab.order == Sviss_Symtab::CONCAT || sviss_symtab.order == Sviss_Symtab::DYNAMIC ); // start with CONCAT ordering when DYNAMIC

  stateSpace_user_cc << endl
		     << "#include <vector>" << endl
		     << endl
		     << "#include \"functions.hh\"" << endl
		     << "#include \"triple.hh\"" << endl
		     << "#include \"utool.hh\"" << endl
		     << "#include \"stateSpace.hh\"" << endl
		     << endl
		     << "const Vector<ushort>& StateSpace::bits(cushort& p, cushort& sbn, const UCudd::Part& part) const {" << endl
		     << "  __SAFE_ASSERT__ (sbn < __sbs.size());" << endl
		     << "  const Symmblock& sb = __sbs[sbn];" << endl
		     << "  __SAFE_ASSERT__ (p < sb.n);" << endl
		     << "  cushort start = bit(p, sbn, 0, part);" << endl
		     << "  return bits(start, sb.local_bits, __local_increment(sbn)); }" << endl
		     << endl
		     << "const UBDD& StateSpace::cube(cushort& p, cushort& sbn, const UCudd::Part& part) const {" << endl
		     << "  __SAFE_ASSERT__ (sbn < __sbs.size());" << endl
		     << "  const Symmblock& sb = __sbs[sbn];" << endl
		     << "  __SAFE_ASSERT__ (p < sb.n);" << endl
		     << "  cushort start = bit(p, sbn, 0, part);" << endl
		     << "  return cube(start, sb.local_bits, __local_increment(sbn)); }" << endl
		     << endl
		     << "cushort StateSpace::bit(cushort& p, cushort& sbn, cushort& b, const UCudd::Part& part) const {" << endl
		     << "  __SAFE_ASSERT__ (sbn < __sbs.size());" << endl
		     << "  const Symmblock& sb = __sbs[sbn];" << endl
		     << "  __SAFE_ASSERT__ (p < sb.n);" << endl
		     << "  __SAFE_ASSERT__ (b < sb.local_bits);" << endl
		     << "  cushort result = sb.offset + " << ( concat ? "p * sb.local_bits * 2 + b * 2" : "b * sb.n * 2 + p * 2" ) << ";" << endl
		     << "  return ( part == UCudd::curr ? result : result + 1 ); }" << endl; }

// Computes stateSpace::__sbs, stateSpace::__global_bits, stateSpace::__N_total
void Parse::finish_symtab() {
  // sanity checks
  __SAFE_ASSERT__ (sb_ns.size() == sb_varCounts.size());
  __SAFE_ASSERT__ (sb_ns.size() == ushort(sb_cnt + 1));
  if (global_varCount == 0 && sb_ns.empty())
    error("finish_symtab: no variables declared");

  Sviss_Symtab::const_iterator record = sviss_symtab.contents.begin();
  Entry::Kind kind;

  // skip parameters and constants
  while (kind = record->entry()->kind, kind == Entry::PARAMETER || kind == Entry::CONSTANT)
    ++record;

  // We first create all symmetry blocks. Then, when we go through global
  // variables and find idst ones, we can add them to the appropriate
  // symmblock.  Also, offset and local_bits of a symmetry block can only be
  // initialized after all symmblocks exist (they might require knowing the
  // sizes of (other) symmblocks).
  stateSpace_user_cc << endl
		     << "void StateSpace::__initialize() {" << endl
		     << "  // setting up symmetry blocks" << endl
		     << "  __sbs.reserve(" << sb_ns.size() << ");" << endl;
  for (ushort sbn = 0; sbn < sb_ns.size(); ++sbn)
    stateSpace_user_cc << "  __sbs.push_back(Symmblock(" << sb_ns[sbn] << ", " << ( sg_ids[sbn] == CLIQUE ? "Symmblock::FULL" : "Symmblock::RING" ) << "));" << endl;
  stateSpace_user_cc << endl;

  // globals
  stateSpace_user_cc << "  // globals idst variables" << endl;
  SymbInt offset = 0;
  SymbInt global_bits = 0;
  for (ushort global = 0; global < global_varCount; ++global, ++record) { // global is just a counter
    Variable* const variable = (Variable*) record->entry();
    __SAFE_ASSERT__ (variable->sbn == -1);
    variable->make_offset(global_bits);
    const SymbInt& numbits = variable->numbits;
    cshort& sb_pointer = variable->sb_pointer;
    check_sb_pointer(sb_pointer, record->symname());
    if (sb_pointer >= 0) {
      const string start_name   = PPRINT::varname("start");
      const string numbits_name = PPRINT::varname("numbits");
      stateSpace_user_cc << "  cushort " << start_name   << " = " << offset  << ";" << endl
			 << "  cushort " << numbits_name << " = " << numbits << ";" << endl
			 << "  __sbs[" << sb_pointer << "].global_idsts.push_back(Triple<ushort>(" << start_name << ", " << numbits_name << ", 2)); "
			 << "// global idst variable '" << record->symname() << "'" << endl; }
    global_bits = global_bits + numbits;
    offset      = offset      + numbits * SymbInt(2); }
  stateSpace_user_cc << "  __global_bits = " << global_bits << ";" << endl;

  SymbInt N = 0;
  for (ushort sbn = 0; sbn < sb_ns.size(); ++sbn) {
    N = N + sb_ns[sbn];
    const string sb_name = "sb_" + PPRINT::widthify(sbn);
    stateSpace_user_cc << endl
		       << "  Symmblock& " << sb_name << " = __sbs[" << sbn << "];" << endl
		       << "  " << sb_name << ".offset = " << offset << ";" << endl;

    // determine sb.local_bits
    SymbInt local_bits = 0;
    bool local_idst_exists = false;
    Sviss_Symtab::const_iterator first_local_record = record; // we need to go through the locals again below
    for (ushort local = 0; local < sb_varCounts[sbn]; ++local, ++record) { // local is just a counter
      Variable* const variable = (Variable*) record->entry();
      __SAFE_ASSERT__ (variable->sbn == sbn);
      cshort& sb_pointer = variable->sb_pointer;
      check_sb_pointer(sb_pointer, record->symname());
      local_idst_exists = local_idst_exists || sb_pointer >= 0;
      variable->make_offset(local_bits);
      local_bits = local_bits + variable->numbits; }

    stateSpace_user_cc << "  " << sb_name << ".local_bits = " << local_bits << ";" << endl;
    offset = offset + sb_ns[sbn] * local_bits * SymbInt(2);

    if (local_idst_exists) {
      // determine Triple for local idst variables. This must wait until sb.local_bits has been computed;
      // thus we must iterate through the local variables twice
      record = first_local_record;
      for (ushort local = 0; local < sb_varCounts[sbn]; ++local, ++record) { // local is just a counter
	const Variable* const variable = (Variable*) record->entry();
	cshort& sb_pointer = variable->sb_pointer;
	if (sb_pointer >= 0)
	  stateSpace_user_cc << "  cushort numbits = " << variable->numbits << ";" << endl
			     << "  for (ushort p = 0; p < " << sb_name << ".n; ++p) {" << endl
			     << "    cushort start = bit(p, " << sbn << ", " << variable->offset << ", UCudd::curr);" << endl
			     << "    __sbs[" << sb_pointer << "].local_idsts[p].push_back(Triple<ushort>(start, numbits, __local_increment(" << sbn << "))); } "
			     << "// local idst variable '" << record->symname() << "'" << endl; }}}

  stateSpace_user_cc << endl
		     << "  __N_total = " << N << ";" << endl
		     << "  ushort numBDDvars = " << offset << ";" << endl // final offset serves as total number of bits
		     << "  ucudd = new UCudd(numBDDvars);" << endl
		     << "  ucudd->Autodyn" << ( sviss_symtab.order == Sviss_Symtab::DYNAMIC ? "Enable(CUDD_REORDER_SIFT)" : "Disable()" ) << ";" << endl
		     << "}" << endl;

  access_functions_and_print(); }

void Parse::finish_APs() {
  R_APs_hpp << endl
	    << "  inline string the_APs() const { return \""; APs.to_stream(R_APs_hpp, ", ") << "\"; }" << endl;

  R_APs_cc << "#include \"functions.hh\"" << endl
	   << "#include \"utool.hh\"" << endl
	   << "#include \"model.hh\"" << endl
	   << "#include \"stateSpace.hh\"" << endl
	   << "#include \"manager.hh\"" << endl
	   << endl;

  for (list<string>::const_iterator ap = APs.begin(); ap != APs.end(); ++ap) {
    const string ap_quotes = string("\"") + (*ap) + "\"";
    R_APs_cc << "UBDD " << (*ap) << "(const StateSpace& S);" << endl
	     << endl
	     << "UBDD Model::the_" << (*ap) << "() {" << endl
	     << "  if (" << (*ap) << ".getNode() == NULL) {" << endl
	     << "    Man.start(" << ap_quotes << ");" << endl
	     << "    " << (*ap) << " = ::" << (*ap) << "(S) & S.States;" << endl
	     << "    Man.end(" << (*ap) << "); }" << endl
	     << "  return " << (*ap) << "; }" << endl
	     << endl; }

  R_APs_cc << "UBDD* Model::AP_pointer(const string& name) {" << endl; // parse needs UBDD*, so return pointer

  size_t width = 0;

  for (list<string>::const_iterator app = APs.begin(); app != APs.end(); ++app)
    width = std::max(width, app->size());

  for (list<string>::const_iterator ap = APs.begin(); ap != APs.end(); ++ap) {
    const string ap_str    = PPRINT::widthify(*ap, width, PPRINT::LEFTJUST);
    const string ap_quotes = PPRINT::widthify(string("\"") + (*ap) + "\"", width + 2, PPRINT::LEFTJUST);
    R_APs_cc << "  if (name == " << ap_quotes << ") { the_" << ap_str << "(); return &" << ap_str << "; }" << endl; }
  R_APs_cc << "  throw CONTROL::Error(\"unknown atomic proposition \" + name); }" << endl; }

void Parse::access_functions_and_print() {
  stateSpace_user_hpp << endl;
  stateSpace_user_cc  << endl;
  std::ostringstream print_header;
  std::ostringstream print_binary;
  std::ostringstream print_decimal;
  std::ostringstream make_States;
  print_header  << endl << "void StateSpace::print_header(ostream& out) const {" << endl;
  print_binary  << endl << "void StateSpace::print_binary(const int* const state, const UCudd::Part& part, ostream& out) const {" << endl;
  print_decimal << endl << "void StateSpace::print_decimal(const int* const state, const UCudd::Part& part, ostream& out) const {" << endl;
  make_States   << endl << "void StateSpace::__make_States() {" << endl
		<<         "  States = UBDD::True();" << endl;
  if (global_varCount > 0) {
    print_header  << "  out << \"(G) \";" << endl;
    print_binary  << "  out << \"(G) \";" << endl;
    print_decimal << "  out << \"Globals:\" << endl;" << endl; }

  Sviss_Symtab::const_iterator record = sviss_symtab.contents.begin();
  Entry::Kind kind;

  // skip parameters and constants in the symbol table
  while (kind = record->entry()->kind, kind == Entry::PARAMETER || kind == Entry::CONSTANT)
    ++record;

  // globals
  for (ushort global = 0; global < global_varCount; ++global, ++record) {
    if (global > 0) {
      print_header << "  out << \"|\";" << endl;
      print_binary << "  out << \"|\";" << endl; }
    const Variable* const variable = (Variable*) record->entry();
    __SAFE_ASSERT__ (variable->scope == Variable::GLOBAL);
    if (variable->type == Variable::RECORD)
      CONTROL::warning("Parse::access_functions_and_print: access functions for record types not defined");
    const string symname  = PPRINT::widthify(record->symname(), sviss_symtab.the_width(), PPRINT::LEFTJUST);
    const SymbInt  start  = variable->offset * SymbInt(2);
    const SymbInt _start  = start + SymbInt(1);
    const SymbInt numbits = variable->numbits;
    const string  body    = "bits(" + string( start) + ", " + string(numbits) + ", 2)";
    const string _body    = "bits(" + string(_start) + ", " + string(numbits) + ", 2)";
    stateSpace_user_hpp << "  inline const vector<ushort>&  "                   << symname << "(                      ) const { return " <<  body << "; }" << endl;
    stateSpace_user_hpp << "  inline const vector<ushort>& "             << "_" << symname << "(                      ) const { return " << _body << "; }" << endl;

    // special: an array has two extra functions for access to individual entries
    if (variable->type == Variable::ARRAY) {
      const Array* const array = (Array*) variable;
      const SymbInt  start  = (variable->offset + SymbInt("i") * array->basevar->numbits) * SymbInt(2);
      const SymbInt _start  = start + SymbInt(1);
      const SymbInt numbits = array->basevar->numbits;
      const string  body    = "bits(" + string( start) + ", " + string(numbits) + ", 2)";
      const string _body    = "bits(" + string(_start) + ", " + string(numbits) + ", 2)";
      stateSpace_user_hpp << "  inline const vector<ushort>&  "                   << symname << "(            cushort& i) const { return " <<  body << "; }" << endl;
      stateSpace_user_hpp << "  inline const vector<ushort>& "             << "_" << symname << "(            cushort& i) const { return " << _body << "; }" << endl; }

    print_header << "  out << \"" << record->symname() << "\";";
    variable->print_binary (print_binary,  variable->offset, 2);
    variable->print_decimal(print_decimal, variable->offset, 2, "\"" + symname + "\"");
    variable->make_States  (make_States,                     2, symname); }

  // symmblocks
  for (ushort sbn = 0; sbn < sb_ns.size(); ++sbn) {
    stateSpace_user_hpp << endl;
    print_header  << endl
		  << "  out << \"   (SB " << sbn << ") \";" << endl;
    print_binary  << endl
		  << "  out << \"   (SB " << sbn << ") \";" << endl
		  << "  for (ushort p = 0; p < " << sb_ns[sbn] << "; ++p) {" << endl
		  << "    if (p > 0)" << endl
		  << "      out << \" # \";" << endl;
    print_decimal << "  out << \"Symmetry block \" << " << sbn << " << \":\" << endl;" << endl
		  << "  for (ushort p = 0; p < " << sb_ns[sbn] << "; ++p) {" << endl
		  << "    out << \"  Process \" << p << \":\" << endl;" << endl;
    make_States   << "  for (ushort p = 0; p < " << sb_ns[sbn] << "; ++p) {" << endl;
    for (ushort local = 0; local < sb_varCounts[sbn]; ++local, ++record) { // local is just a counter
      if (local > 0) {
	print_header << "    out << \"|\";" << endl;
	print_binary << "    out << \"|\";" << endl; }
      const Variable* const variable = (Variable*) record->entry();
      __SAFE_ASSERT__ (variable->scope == Variable::LOCAL);
      __SAFE_ASSERT__ (variable->sbn   == sbn);
      if (variable->type == Variable::RECORD)
	CONTROL::warning("Parse::access_functions_and_print: access functions for record types not defined");
      const string symname = PPRINT::widthify(record->symname(), sviss_symtab.the_width(), PPRINT::LEFTJUST);
      const string start_init = "bit(p, " + PPRINT::widthify(sbn) + ", " + string(variable->offset);
      const SymbInt numbits = variable->numbits;
      const string  start = start_init + ", UCudd::curr)";
      const string _start = start_init + ", UCudd::next)";
      stateSpace_user_hpp << "  inline const vector<ushort>&  "                   << symname << "(cushort& p            ) const { return bits(" <<  start << ", " << numbits << ", __local_increment(" << sbn << ")); }" << endl;
      stateSpace_user_hpp << "  inline const vector<ushort>& "             << "_" << symname << "(cushort& p            ) const { return bits(" << _start << ", " << numbits << ", __local_increment(" << sbn << ")); }" << endl;

      // special: access functions for array: see globals
      if (variable->type == Variable::ARRAY) {
	const Array* const array = (Array*) variable;
	const string start_init = "bit(p, " + PPRINT::widthify(sbn) + ", " + string(variable->offset + SymbInt("i") * array->basevar->numbits);
	const SymbInt numbits = array->basevar->numbits;
	const string  start = start_init + ", UCudd::curr)";
	const string _start = start_init + ", UCudd::next)";
	stateSpace_user_hpp << "  inline const vector<ushort>&  "                   << symname << "(cushort& p, cushort& i) const { return bits(" <<  start << ", " << numbits << ", __local_increment(" << sbn << ")); }" << endl;
	stateSpace_user_hpp << "  inline const vector<ushort>& "             << "_" << symname << "(cushort& p, cushort& i) const { return bits(" << _start << ", " << numbits << ", __local_increment(" << sbn << ")); }" << endl; }

      print_header << "  out << \"" << record->symname() << "\";";
      variable->print_binary (print_binary , variable->offset, 4);
      variable->print_decimal(print_decimal, variable->offset, 4, "\"" + symname + "\"");
      variable->make_States  (make_States,                     4, symname); }
    print_header  <<          endl;
    print_binary  << "  }" << endl;
    print_decimal << "  }" << endl;
    make_States   << "  }" << endl; }

  print_header  << "}" << endl;
  print_binary  << "}" << endl;
  print_decimal << "}" << endl;
  make_States   << "  _States = States.shift_to_next();" << endl
		<< "}" << endl;

  stateSpace_user_cc << print_header .str();
  stateSpace_user_cc << print_binary .str();
  stateSpace_user_cc << print_decimal.str();
  stateSpace_user_cc << make_States  .str(); }

// Note: if symname points to a constant the symbol table value of which is
// not a number literal, then we return the name of the constant, NOT the
// expression contained for it in the symtab. This prevents multiple
// evaluations of the same expression.
// Example: assume n is a parameter, and
// const l = n + 3
// const v = l
// Should v get the symbolic value "l" or "n + 3"? I think the former, since otherwise n + 3 is evaluated twice
SymbInt Parse::value_of_param_or_const(const string& symname) {
  const Entry* const & entry = sviss_symtab.lookup(symname);
  if      (entry->kind == Entry::PARAMETER)
    return symname;
  else if (entry->kind == Entry::CONSTANT) {
    Constant* constant = (Constant*) entry;
    return ( constant->value.is_pos_literal() ? constant->value : SymbInt(symname) ); } // see above
  else
    return error("value_of_param_or_const: parameter or constant expected; found '" + symname + "'"), ""; }

void Parse::variable_decl(const string& symname, Variable* variable) {
  switch(variable->type) {
  case Variable::ENUMERATE: enumerate_decl(symname, variable); break;
  case Variable::RANGE:         range_decl(symname, variable); break;
  case Variable::ARRAY:         array_decl(symname, variable); break;
  case Variable::RECORD:       record_decl(symname, variable); break;
  default: error("variable_decl: internal: wrong case"); }}

void Parse::enum_typedef(const string& symname, Enumerate* enumerate) {
  if (! enumerate->is_boolean()) { // no duplicate check or typedefs for "bool"
    const List<string>& values = enumerate->values;
    List<string>::const_iterator duplicate = values.duplicate();
    if (duplicate != values.end())
      error("enumerate_decl: value list has duplicate element: " + *duplicate);
    stateSpace_user_hpp << "  struct " << symname << " { enum { " << values << " }; };" << endl; }}

void Parse::enumerate_decl(const string& symname, Variable* variable) {
  Enumerate* enumerate = (Enumerate*) variable;
  if (! enumerate->is_boolean())
    enum_typedef(symname, enumerate);
  sviss_symtab.insert_variable(symname, variable); }

// range l..r a:
// when referenced, replace a by a+l
// when assigned,   replace the right-hand side exp by exp-l
// Examples:
// if (a==6)  becomes if (a==(6+l))
// a := 10    becomes a := 10 - r
// a := a + 1 becomes a := ((a+l)+1)-l. How do we optimize that to a := a + 1? What about a := a + b for two range variables a, b?
//
// A much easier solution is of course to ignore the lower limit when initializing variables,
// and to only use it when assignments, etc are checked ("non-compressed"). This is a waste of space, but may save access time.
void Parse::range_decl(const string& symname, Variable* variable) {
  const Range* range = (Range*) variable;
  const SymbInt& left  = range->left;
  // const SymbInt& right = range->right;
  if (left != 0)
    CONTROL::warning("Parse::range_decl: left boundary " + string(left) + " for range variable " + symname + " will be extended down to zero");
  sviss_symtab.insert_variable(symname, variable); }

void Parse::array_decl(const string& symname, Variable* variable) {
  sviss_symtab.insert_variable(symname, variable);
  const Variable* basevar = ((Array*) variable)->basevar;
  if (basevar->type == Variable::ENUMERATE)
    enum_typedef(symname, (Enumerate*) basevar); } // for array[] of enum, we make symbolic enum constants using the array variable name

void Parse::record_decl(const string& symname, Variable* variable) {
  sviss_symtab.insert_variable(symname, variable); }

void Parse::check_sb_pointer(cshort& sb_pointer, const string& symname) {
  if (sb_pointer < -1 || sb_pointer >= short(sb_ns.size()))
    error("check_sb_pointer: idst variable " + symname + " points to non-existent symmetry block " + PPRINT::widthify(sb_pointer)); }

void Parse::AP_decl(const string& symname) {
  if (std::find(APs.begin(), APs.end(), symname) != APs.end())
    error("AP_decl: atomic proposition " + symname + " declared twice");
  APs.push_back(symname);
  R_APs_hpp << "  UBDD " << symname << ";" << endl
	    << "  UBDD the_" << symname << "();" << endl; }

void Parse::open_verbose(FILE*& f, const string& fname) {
  if (verbose)
    cerr << "opening file " << fname << " for reading" << endl;
  if ( (f = fopen(fname.c_str(), "r")) == NULL)
    error("open_verbose: can't open file " + fname + " for reading"); }

void Parse::open_verbose(ofstream& f, const string& fname) {
  if (verbose)
    cerr << "opening file " << fname << " for writing" << endl;
  f.open(fname.c_str());
  if (!f.is_open())
    error("open_verbose: can't open file " + fname + " for writing"); }

void Parse::close_verbose(FILE* f, const string& fname) {
  if (fclose(f) != 0)
    error("close_verbose: can't close file " + fname);
  if (verbose)
    cerr << "file " << fname << " closed." << endl; }

void Parse::close_verbose(ofstream& f, const string& fname) {
  f.close();
  if (f.is_open())
    error("close_verbose: can't close file " + fname);
  if (verbose)
    cerr << "file " << fname << " closed." << endl; }

void Parse::remove_verbose(FILE* f, const string& fname) {
  if (fclose(f) != 0)
    error("remove_verbose: can't close file " + fname);
  remove(fname.c_str());
  cerr << "file " << fname << " removed." << endl; }

void Parse::remove_verbose(ofstream& f, const string& fname) {
  f.close();
  if (f.is_open())
    error("remove_verbose: can't close file " + fname);
  remove(fname.c_str());
  cerr << "file " << fname << " removed." << endl; }
