#ifndef __PARSE_HH__
#define __PARSE_HH__

#include <stdio.h>

#include <string>
#include <ostream>
#include <fstream>
#include <sstream>
#include <list>
#include <vector>
#include <set>

#include "sviss-symtab.hh"
#include "symbint.hh"
#include "heap.hh"

using std::string;
using std::ifstream;
using std::ofstream;

struct Parse {
  string ifilename;

  FILE* ifile; // cannot use ifstream here since yyin is a FILE*

  // _cpp are non-selfsufficient files copied into another C++ source file, _hpp are non-selfsufficient files copied into another C++ header file,
  // _cc are self-sufficient C++ source files (to be compiled separately).
  // These distinctions are important for an efficient Makefile
  ofstream run_parameter_cpp;
  ofstream run_construct_stateSpace_cpp;
  ofstream stateSpace_user_hpp;
  ofstream stateSpace_user_cc;
  ofstream R_APs_hpp;
  ofstream R_APs_cc;

  bool verbose;

  static const Set<string> RESERVED;

  ushort line_number; // within the sviss source file

  typedef enum { CLIQUE, RING } Symmtype;

  ushort           global_varCount; // number of global variables
  short            sb_cnt;          // number of current symmblock (-1 if current variable global)
  vector<SymbInt>  sb_ns;           // number of processes in symmblocks
  vector<Symmtype> sg_ids;          // symmetry groups in symmblocks (clique, ring)
  vector<ushort>   sb_varCounts;    // number of variables in symmblocks
  List<string>     APs;             // names of atomic propositions

  typedef Heap<     string  > S_Heap;
  typedef Heap<     SymbInt > I_Heap;
  typedef Heap<list<string> > L_Heap;
  // remember that the Entry* created during parsing are deleted when the symbol table is destroyed. No heap.

  S_Heap s_heap; // string* that are to be deleted at end
  I_Heap i_heap; // symbolic integers
  L_Heap l_heap; // list<string>*

  Sviss_Symtab sviss_symtab;

  string parameter_decls;
  string parameter_inits;
  string parameter_passes;

  string constant_inits;

  Parse(const string& ifilename, const string& verbose = "true");
  ~Parse();

  void print_symtab(ostream& out = cout) const;

  void   open_verbose(   FILE*& f, const string& fname);
  void   open_verbose(ofstream& f, const string& fname);
  void  close_verbose(   FILE*  f, const string& fname);
  void  close_verbose(ofstream& f, const string& fname);
  void remove_verbose(   FILE*  f, const string& fname);
  void remove_verbose(ofstream& f, const string& fname);
  void remove_verbose(); // all

  inline void error(const string& message) { remove_verbose(); throw CONTROL::Error("Parse::" + message); }

  inline void check_reserved_words(const string& identifier) {
    if (RESERVED.count(identifier))
      error("check_reserved_words: identifier '" + identifier + "' is a reserved word"); }

  void parameter_decl(const string& symname, const string& meaning = "\"\"", const string& default_value = "\"\"");
  void  constant_decl(const string& symname, const SymbInt& value);
  void  variable_decl(const string& symname, Variable* variable);

  void enumerate_decl(const string& symname, Variable* variable);
  void     range_decl(const string& symname, Variable* variable);
  void     array_decl(const string& symname, Variable* variable);
  void    record_decl(const string& symname, Variable* variable);

  void enum_typedef  (const string& symname, Enumerate* enumerate);

  void AP_decl       (const string& symname);

  SymbInt value_of_param_or_const(const string& symname);

  void check_sb_pointer(cshort& sb_pointer, const string& symname);

  void stateSpace();
  void finish_symtab();
  void finish_APs();

  void access_functions_and_print();

  // Variables occur in pairs: current + next_state. Since we do not allow identifiers
  // in input files to begin with _, the below is guaranteed to create a new name
  static inline string next_state_name(const string& name) { return "_" + name; } };

#endif
