#include <stdio.h>

#include <iostream>
#include <string>
#include <cctype>

#include "functions.hh"
#include "input.hh"
#include "utool.hh"
#include "stateSpace.hh"
#include "manager.hh"
#include "model.hh"
#include "model-check.hh"
#include "model-check-reduced.hh"

UCudd* ucudd = NULL;

// From Lex
struct yy_buffer_state;
typedef struct yy_buffer_state* YY_BUFFER_STATE;
YY_BUFFER_STATE yy_scan_string(const char*);
void yy_delete_buffer(YY_BUFFER_STATE);

// From Yacc
extern int yyparse(ModelCheck*, Model&);

// If spec is empty, only print statistics
void run(const StateSpace& S, Manager& Man, Model& M, const string& spec = "", const string& alg = "NOSYMM", const bool& iter = false) {
  // MC MUST be a pointer. If declared ModelCheck, it will not be interpreted dynamically as ModelCheck_reduced, even if passed as reference
  static ModelCheck*         MC;
  static ModelCheck*         MC_nosymm = NULL;
  static ModelCheck_reduced* MC_red    = NULL;

  if (alg == "NOSYMM") {
    if (MC_nosymm == NULL)
      MC_nosymm = new ModelCheck(S, Man, M, iter);
    MC = MC_nosymm; }
  else {
    if (MC_red == NULL)
      MC_red = new ModelCheck_reduced(S, Man, M, iter);
    MC_red->set_algorithm(alg);
    MC = MC_red; }

  if (! spec.empty()) {
    YY_BUFFER_STATE buffer = yy_scan_string(spec.c_str());
    yyparse(MC, M);
    yy_delete_buffer(buffer); }
  Man.print(); }


int main(const int argc, const char * const * const argv) {

  try {

    Input I;

    // include parameters, such as number of processes
# include "run-parameter.cpp"

    I.add_switch  ("-s",    "print statespace information (mainly internal use)");

    I.add_argument("-spec", "specification; use '' if contains spaces; if not given, you will be prompted later", "", "-");
    I.add_argument("-alg",  "reduction algorithm", "NOSYMM,ORBIT,MULTIPLE,DYNAMIC", "NOSYMM"); // __PARTITIONED__: PARTITIONED
    I.add_switch  ("-iter", "print statistics during iterations of fixpoint computations");

    try { I.get_command_line(argc, argv); }
    catch (Input::Help) { return 0; }

    const bool   ss_info = ( I.arg_value("-s")    == "true" );
    const string spec    =   I.arg_value("-spec");
    const string alg     =   I.arg_value("-alg");
    const bool   iter    = ( I.arg_value("-iter") == "true" );

    if (ss_info) {
      cout << endl;
      I.print_command_line(); }

    // Build state space
# include "run-construct-stateSpace.cpp"
    if (ss_info) {
      cout << endl;
      cout << "State space information:" << endl;
      S.__print(); }

    Manager Man(S);
    Model   M  (S, Man);

    if (spec != "-") {
      run(S, Man, M, spec, alg, iter);
      return 0; }

    Input J;
    J.add_argument("-alg", "reduction algorithm", "NOSYMM,ORBIT,MULTIPLE,DYNAMIC", "NOSYMM"); // __PARTITIONED__: PARTITIONED
    J.add_switch  ("-iter", "print statistics during iterations of fixpoint computations");

    for (;;) {
      string args;
      cout << "available atomic propositions: " << M.the_APs() << endl;
      cout << "verification task (format: \"<spec> -alg <algorithm> -iter\" or -h for help): ";
      std::getline(cin, args);
      string args_copy = args;
      string::iterator first_ws = std::remove_if(args_copy.begin(), args_copy.end(), isspace);
      args_copy.erase(first_ws, args_copy.end()); // erase whitespace

      if (args_copy.empty()) {
	run(S, Man, M);
	continue; }
      if (args_copy == "-h") {
	cout << "  format: <spec> -alg <algorithm> -iter" << endl
	     << "  All except <spec> are optional" << endl
	     << "  <spec> = CTL specification, e.g. Bad & FS Init. No \" or ' required, but <spec> must come first" << endl
	     << "  -alg   = reduction algorithm, possible values { NOSYMM ORBIT MULTIPLE DYNAMIC }, default 'NOSYMM' or previous alg" << endl // __PARTITIONED__: PARTITIONED
	     << "  -iter  = print statistics during iterations of fixpoint computations, default false" << endl
	     << "  NOSYMM, etc, may be abbreviated (N for NOSYMM, O for ORBIT, ...)" << endl;
	continue; }

      string::size_type first_dash = args.find('-');
      const string spec = ( first_dash == string::npos ? args : args.substr(0, first_dash) );
      args = ( first_dash == string::npos ? "" : args.substr(first_dash) );

      J.get_command_line("", args);
      // J.get_command_line("", args, Input::SKIP_USAGE);

      const string alg  =   J.arg_value("-alg");
      const bool   iter = ( J.arg_value("-iter") == "true" );

      run(S, Man, M, spec, alg, iter);

      J.reset_switches(); }

    return 0; }

  // error handling

  catch (const CONTROL::Error& error) { error.print_exit(); }}
