
#ifndef __STATE_SPACE__
#define __STATE_SPACE__

#include <assert.h>

#include <string>
#include <vector>
#include <iostream>

#include "functions.hh"
#include "triple.hh"

#include "utool.hh"


struct Symmblock {

  typedef enum { FULL, RING } Symmgroup;

  cushort n;           // number of processes represented in this block
  Symmgroup symmgroup; // processes form a clique (full symmetry) or a ring (rotational symmetry)

  ushort offset;       // number of bits preceding the bits in this block
  // ushort l;         // number of local states of the processes represented in this block
  ushort local_bits;   // number of bits per process in this block. Curr-state only. Roughly (but not quite) equal to ceil_log(l, 2). Used by specifics.

  // id-sensitive variables are stored as triples (start, numbits, increment). When needed, the bit vector is retrieved using curr/next_bits.
          Vector<Triple<ushort> >   global_idsts; // global id-sensitive variables associated with this block
  Vector< Vector<Triple<ushort> > >  local_idsts; // one vector per process

  inline Symmblock(cushort& n, const Symmgroup& symmgroup):
    n(n), symmgroup(symmgroup), local_idsts(n) {} // offset and local_bits are initialized later; see parse.cc:initialize for reason
  inline Symmblock& operator=(const Symmblock& sb) { assert (false); return *this; } // copy non-const members. We don't need this, but the compiler wants it

  void print(cushort& indent = 0, ostream& out = cout) const; };

using LOGARITHM::ceil_log; // may be used in stateSpace-user.hpp included below. Can not be declared within class
using LOGARITHM::power;

// In this class, all identifiers for data members, and all identifiers for methods that can be called with 0 arguments or
// with one ushort argument begin with __ to prevent nameclashes with user definitions
// (name X in stateSpace-user.hpp creates methods X(.) and _X(.), with <= 1 argument)
// Plus, the names States, invariant, invariant_but, invariant_sbn, invariant_all are reserved words.
// Remember to keep the list of reserved words uptodate in Parse::RESERVED
struct StateSpace {
  ushort            __global_bits; // number of bits for global variables. Curr-state only
  Vector<Symmblock> __sbs;
  ushort            __N_total;     // total number of processes (in all blocks)

  void __make_States();
  inline ushort __local_increment(cushort& sbn) const {  // difference b/w consecutive bits within one process
    return ( __sbs[sbn].local_bits == 1 ? 0 :  // local increment doesn't matter in this case, but the diff trick below will give an error (no bit 1)
	     bit(0, sbn, 1, UCudd::curr) -
	     bit(0, sbn, 0, UCudd::curr) ); } // valid for all orders!

  // these functions are used by the parser when it maps variable expressions to bit vectors
  static const Vector<ushort>& bits(cushort& start, cushort& numbits, cushort& increment);
  static const UBDD&           cube(cushort& start, cushort& numbits, cushort& increment);

  // these functions are used by model-check-reduced, which knows (idst) variables only in the Triple form
  static inline const Vector<ushort>& bits(const Triple<ushort>& var, const UCudd::Part& part) { return bits(var.first + ( part == UCudd::curr ? 0 : 1 ), var.second, var.third); }
  static inline const UBDD&           cube(const Triple<ushort>& var, const UCudd::Part& part) { return cube(var.first + ( part == UCudd::curr ? 0 : 1 ), var.second, var.third); }

  // all bits of a process as a vector or a cube; individual bit of a process
  const Vector<ushort>& bits(cushort& p, cushort& sbn,             const UCudd::Part& part) const;
  const UBDD&           cube(cushort& p, cushort& sbn,             const UCudd::Part& part) const;
  cushort               bit (cushort& p, cushort& sbn, cushort& b, const UCudd::Part& part) const;

  void __print(cushort& indent = 0, ostream& out = cout) const;
  void print_header (ostream& out = cout) const;
  void print_binary (const int* const state, const UCudd::Part& part, ostream& out = cout) const;
  void print_decimal(const int* const state, const UCudd::Part& part, ostream& out = cout) const;

  void print_binary (const int* const state, cushort& start, cushort& numbits, cushort& increment, ostream& out = cout) const; // start, start + increment, ...
  long decimal      (const int* const state, cushort& start, cushort& numbits, cushort& increment) const;

  void __initialize(); // initializes __sbs, __N_total

  // Library functions for the user
  UBDD  States;                                         // the real state space, taking variable ranges into consideration. Needed for CEX creation
  UBDD _States;                                         // the real state space, next-state version (not intended for user)
  UBDD invariant    (cushort& p,   cushort& sbn) const; // curr and next local state of p are same
  UBDD invariant_but(cushort& p,   cushort& sbn) const; // all processes in the given symmblock but p are invariant
  UBDD invariant_sbn(cushort& sbn)               const; // all processes in the given symmblock are invariant
  UBDD invariant_all()                           const; // all processes invariant (only globals change)
  UBDD invariant_globals()                       const; // global variables invariant

# include "stateSpace-user.hpp"

};

// The following structs are used with enumerate functions and always print with offset 2
struct Print_curr_S: public Read_bits {
  const StateSpace& S;
  ostream& out;
  inline Print_curr_S(const StateSpace& S, ostream& out = cout): S(S), out(out) {}
  inline void operator()(const int* const state) { out << "  "; S.print_binary(state, UCudd::curr, out); out << endl; } };

struct Print_next_S: public Read_bits {
  const StateSpace& S;
  ostream& out;
  inline Print_next_S(const StateSpace& S, ostream& out = cout): S(S), out(out) {}
  inline void operator()(const int* const state) { out << "  "; S.print_binary(state, UCudd::next, out); out << endl; } };

struct Print_curr_next_S: public Read_bits {
  const StateSpace& S;
  ostream& out;
  inline Print_curr_next_S(const StateSpace& S, ostream& out = cout): S(S), out(out) {}
  inline void operator()(const int* const state) {
    out << "  ";      S.print_binary(state, UCudd::curr, out);
    out << "  ==>  "; S.print_binary(state, UCudd::next, out); out << endl; } };

#endif
