
#ifndef __SVISS_SYMTAB__
#define __SVISS_SYMTAB__

#include <string>
#include <list>
#include <ostream>

#include "functions.hh"
#include "symbint.hh"
#include "symtab.hh"

using std::string;
using std::ostream;

// An "entry" is the information associated with a symbol name. Its
// contents depends on whether the symbol stands for a parameter, a
// constant, or a program variable.
struct Entry {
  typedef enum { PARAMETER, CONSTANT, VARIABLE } Kind; // we sometimes have to know the kind of entry, for casts
  const Kind kind;
  string the_kind() const; // returns "parameter", "constant" or "variable"
  inline Entry(const Kind& kind): kind(kind) {}
  virtual inline void print(cushort& indent = 0, ostream& out = cout) const { out << the_kind(); } // by default just print what you are
  virtual ~Entry() {} };

// A parameter has no data members
struct Parameter: public Entry {
  const string meaning;
  const string default_value;
  Parameter(const string& meaning = "\"\"", const string& default_value = "\"\""): Entry(PARAMETER), meaning(meaning), default_value(default_value) {}
  void print(cushort& indent = 0, ostream& out = cout) const; };

struct Constant: public Entry {
  const SymbInt value;
  inline Constant(const SymbInt& value): Entry(CONSTANT), value(value) {}
  inline void print(cushort& indent = 0, ostream& out = cout) const { out << "constant of value " << value; } };

// A "Variable" is everything you think of a program variable, except a name. That's why this type can also
// be used to represent unnamed variables, as the 10 variables of type bool in "var x:array[10] of bool".
// Variables only have names at the top level (name stored in the symbol table) and in records (stored there)
struct Variable: public Entry {
  typedef enum { GLOBAL, LOCAL }                   Scope;
  typedef enum { ENUMERATE, RANGE, ARRAY, RECORD } Type;
  const Scope scope;
  const Type  type;
  inline string the_scope() const { return scope == GLOBAL ? "global" : "local"; }
  string the_type () const;

  SymbInt offset;                  // for global: offset from 0. For local: offset from start of symmblock. All offsets in bits. Irrespective of variable order
  SymbInt numbits;                 // number of bits to represent the variable (additive, unlike size!). Non-const since can only initialize in subclass
  cshort sbn;                      // if global, -1. If local, number of symmetry block to which variable belongs
  cshort sb_pointer;               // if not id-sensitive, -1. If id-sensitive, the symmetry block the variable points to
  virtual inline void make_offset(const SymbInt& offset) { this->offset = offset; }
  virtual        void make_numbits() = 0; // called in constructor of subclass

  inline Variable(const Type& type, cshort& sbn, cshort& sb_pointer = -1):
    Entry(VARIABLE), scope ( sbn == -1 ? GLOBAL : LOCAL ), type(type), offset("undefined"), numbits("undefined"), sbn(sbn), sb_pointer(sb_pointer) {}
  virtual void print(cushort& indent = 0, ostream& out = cout) const; // must redeclare here since it is redefined here
  virtual void print_binary (ostream& s, const SymbInt& offset, cushort& indent                    ) const; // offset = total offset (so far)
  virtual void print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const; // offset = total name   (so far)
  virtual void make_States  (ostream& s,                        cushort& indent, const string& name) const = 0;
  virtual inline ~Variable() {} };

struct Enumerate: public Variable {
  const List<string> values;
  static List<string> bool_identifiers; // { false, true }. Used to represent a boolean as special ENUMERATE type without having to make a subtype
  inline void make_numbits() { numbits = LOGARITHM::ceil_log(values.size(), 2); }
  inline Enumerate(cshort& sbn, const List<string>& values = bool_identifiers): // if no values given, "false, true" is assumed
    Variable(ENUMERATE, sbn), values(values, ", ") { make_numbits(); }
  inline bool is_boolean() const { return values == bool_identifiers; }
  inline void print(cushort& indent = 0, ostream& out = cout) const { Variable::print(); out << ", values { " << values << " }"; }
  void print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const; 
  void make_States  (ostream& s,                        cushort& indent, const string& name) const; };

struct Range: public Variable {
  const SymbInt left;
  const SymbInt right;
  inline void make_numbits() { numbits = funcall("ceil_log", right + SymbInt(1), 2); } // right-left+1 vs. right+1!
  inline Range(cshort& sbn, const SymbInt& left, const SymbInt& right):
    Variable(RANGE, sbn), left(left), right(right) { make_numbits(); }
  Range(cshort& sbn, const SymbInt& sb_pointer); // for id-sensitive variables
  void print(cushort& indent = 0, ostream& out = cout) const;
  void make_States(ostream& s, cushort& indent, const string& name) const; };

struct Array: public Variable {
  const SymbInt size; // number of components in the array
  Variable* const basevar; // not const, since its offset is set later
  inline void make_offset(const SymbInt& offset) { Variable::make_offset(offset); basevar->make_offset(offset); } // basevar has same offset as array var
  inline void make_numbits() { numbits = size * basevar->numbits; }
  inline Array(cshort& sbn, const SymbInt& size, Variable* const basevar):
    Variable(ARRAY, sbn), size(size), basevar(basevar) { make_numbits(); }
  inline void print(cushort& indent = 0, ostream& out = cout) const { Variable::print(); out << ", size " << size << ", basetype "; basevar->print(indent); } // must pass indent here!
  void print_binary (ostream& s, const SymbInt& offset, cushort& indent) const;
  void print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const;
  void make_States  (ostream& s, cushort& indent, const string& name) const;
  inline ~Array() { delete basevar; } };

// A record has its own symbol table. make_numbits is not called in the constructor (as it is with the other types), since
// information about members of symtab is not yet available at construction time (incremental insert!). Instead, it is called
// in sviss.y when the parsing of the record is complete.
struct Record: public Variable {
  Symtab<Variable*> symtab;
  void make_offset(const SymbInt& offset);
  void make_numbits();
  inline Record(cshort& sbn): Variable(RECORD, sbn) {}
  inline void insert(const string& symname, Variable* const v) { symtab.insert(symname, v); } // for now without scope
  void print(cushort& indent = 0, ostream& out = cout) const;
  void print_binary (ostream& s, const SymbInt& offset, cushort& indent) const;
  void print_decimal(ostream& s, const SymbInt& offset, cushort& indent, const string& name) const;
  void make_States  (ostream& s,                        cushort& indent, const string& name) const;
  ~Record(); };

struct Sviss_Symtab: public Symtab<Entry*> {

  typedef enum { CONCAT, INTERL, DYNAMIC } Order;
  Order order;
  string the_order() const;

  inline void insert_parameter(const string& symname, Parameter* parameter) { insert(symname, parameter); }
  inline void insert_constant (const string& symname, Constant * constant ) { insert(symname, constant ); }
  inline void insert_variable (const string& symname, Variable * variable ) { insert(symname, variable ); }

  inline Sviss_Symtab(): order(CONCAT) {}
  void print(cushort& indent = 0, ostream& out = cout) const;
  ~Sviss_Symtab(); };

#endif
