%{

// Parsing SVISS descriptions

#include <stdlib.h>
#include <assert.h>
#include <string>
#include <iostream>
#include <fstream>
#include <list>
#include <utility>

#include "sviss-symtab.hh"
#include "parse.hh"
#include "functions.hh"
#include "input.hh"
#include "symbint.hh"
#include "symtab.hh"

#define YYLEX_PARAM parse

// From lex
extern int yylex(Parse& parse);

void yyerror(Parse& parse, char const *msg);

using std::string;
using std::cout;
using std::cerr;
using std::endl;

%}

%error-verbose

%parse-param {Parse& parse}

%union {
  ushort                       num;        // for numeric values, as in [0..2]
  string*                      str;        // for ascii strings: identifier names, text
  SymbInt*                     symbint;    // for symbolic integers: expressions
  Variable*                    variable;   // for type declarations
  std::pair<string,Variable*>* varRecord;  // for individual record declaration
  Record*                      record;     // for record declaration
  list<string>*                id_list; }; // for lists of identifiers

// terminal symbols
%token PARAMETER_TOK CONST_TOK ORDER_TOK GLOBAL_TOK CLIQUE_TOK RING_TOK AP_TOK CODE_TOK // top level
%token CONCAT_TOK INTERL_TOK DYNAMIC_TOK                                                // variable order
%token ENUM_TOK RANGE_TOK BOOL_TOK ARRAY_TOK OF_TOK RECORD_TOK                          // type
%token <num> NUM_TOK                                                                    // number
%token <str> ID_TOK                                                                     // identifier
%token <str> TEXT_TOK                                                                   // text
%token <str> CEIL_LOG_TOK POWER_TOK                                                     // floating point function
%token SB_TOK                                                                           // symmetry block related tokens
%nonassoc ASSGN_TOK DOTDOT_TOK ATM_TOK ATL_TOK NEQ_TOK                                  // character combination

// non-terminal symbols
%type <num>       symmgroup_identifier
%type <symbint>   expr
%type <variable>  type_decl
%type <varRecord> record_decl
%type <record>    record_decls
%type <id_list>   id_list

// arithmetic
%left '+' '-'
%left '*' '/'

%%

sviss: parameters { parse.stateSpace_user_hpp << endl; } // parameters first, since constants may depend on them (but not vice versa)
       constants
       order      { parse.stateSpace(); }
       globals
       symmblocks { parse.finish_symtab(); }
       APs        { parse.finish_APs(); }
       CODE_TOK   // this serves merely to end the AP section of the parser. The lexer takes it from here. This is not optional since R must be defined


parameters: /* empty */
  |         PARAMETER_TOK parameter_decls ';'

parameter_decls: parameter_decl
  |              parameter_decls ';' parameter_decl

parameter_decl: ID_TOK                               { parse.parameter_decl(*$1          ); }
  |             ID_TOK '(' TEXT_TOK ')'              { parse.parameter_decl(*$1, *$3     ); }
  |             ID_TOK '(' TEXT_TOK ',' TEXT_TOK ')' { parse.parameter_decl(*$1, *$3, *$5); }


constants: /* empty */
  |        CONST_TOK constant_decls

constant_decls: constant_decl
  |             constant_decls constant_decl

constant_decl: ID_TOK '=' expr ';' { parse.constant_decl(*$1, *$3); }


order: /* empty */
  |    ORDER_TOK order_spec ';'

order_spec: CONCAT_TOK  { parse.sviss_symtab.order = Sviss_Symtab::CONCAT;  }
  |         INTERL_TOK  { parse.sviss_symtab.order = Sviss_Symtab::INTERL;  }
  |         DYNAMIC_TOK { parse.sviss_symtab.order = Sviss_Symtab::DYNAMIC; }


globals: /* empty */
  |      GLOBAL_TOK global_decls

global_decls: global_decl              { ++parse.global_varCount; }
  |           global_decls global_decl { ++parse.global_varCount; }

// ID_TOK is repeated since we need it in the call to parse.variable_decl
global_decl: ID_TOK ':' type_decl           ';' { parse.variable_decl(*$1, $3); }
  |          ID_TOK ':' SB_TOK '(' expr ')' ';' { parse.variable_decl(*$1, new Range(-1, (*$5))); }


symmblocks: /* empty */
  |         symmblocks symmblock

symmblock: symmgroup_identifier '[' expr ']' { ++parse.sb_cnt; parse.sg_ids.push_back(Parse::Symmtype($1));
	                                       parse.sb_ns.push_back(*$3); parse.sb_varCounts.push_back(0); } local_decls // new symmblock

symmgroup_identifier: CLIQUE_TOK { $$ = Parse::CLIQUE; }
  |                     RING_TOK { $$ = Parse::RING; }

local_decls: local_decl              { ++parse.sb_varCounts.back(); }
  |          local_decls local_decl  { ++parse.sb_varCounts.back(); }

local_decl: ID_TOK ':' type_decl           ';' { parse.variable_decl(*$1, $3); }
//  |         ID_TOK ':' SB_TOK '(' expr ')' ';' { parse.variable_decl(*$1, new Range(parse.sb_cnt, (*$5))); } // not yet implemented


record_decls: record_decl              { $$ = new Record(parse.sb_cnt); $$->insert($1->first, $1->second); delete $1; } // must be recursive in FIRST argument,
  |           record_decls record_decl { $$ = $1;                       $$->insert($2->first, $2->second); delete $2; } // to make symtab order = declaration order

record_decl: ID_TOK ':' type_decl ';'  { $$ = new std::pair<string, Variable*>(*$1, $3); }


type_decl: BOOL_TOK                                { $$ = new Enumerate(parse.sb_cnt); }
  |        ENUM_TOK '{' id_list '}'                { $$ = new Enumerate(parse.sb_cnt, *$3); }
  |        RANGE_TOK expr DOTDOT_TOK expr          { $$ = new Range    (parse.sb_cnt, *$2, *$4); }
  |        RANGE_TOK '[' expr ']'                  { $$ = new Range    (parse.sb_cnt, 0, (*$3) - SymbInt(1)); }
  |        ARRAY_TOK '[' expr ']' OF_TOK type_decl { $$ = new Array    (parse.sb_cnt, *$3, $6); }
  |        RECORD_TOK '{' record_decls '}'         { $$ = $3; $$->make_numbits(); }  // see note in sviss-symtab.hh about make_numbits


id_list: ID_TOK              { $$ = new list<string>(1, *$1); parse.l_heap.insert($$); }
  |      id_list ',' ID_TOK  { $$ = $1; $$->push_back(*$3); }  // reuse $1

// 'expr' is an expression that depends only on parameters or constants, not on program variables
expr: NUM_TOK                            { $$ = new SymbInt($1);                                 parse.i_heap.insert($$); }
expr: ID_TOK                             { $$ = new SymbInt(parse.value_of_param_or_const(*$1)); parse.i_heap.insert($$); }
expr: expr '+' expr                      { $$ = new SymbInt((*$1) + (*$3));                      parse.i_heap.insert($$); }
expr: expr '-' expr                      { $$ = new SymbInt((*$1) - (*$3));                      parse.i_heap.insert($$); }
expr: expr '*' expr                      { $$ = new SymbInt((*$1) * (*$3));                      parse.i_heap.insert($$); }
expr: expr '/' expr                      { $$ = new SymbInt((*$1) / (*$3));                      parse.i_heap.insert($$); }
expr: CEIL_LOG_TOK '(' expr ',' expr ')' { $$ = new SymbInt(funcall("ceil_log", *$3, *$5));      parse.i_heap.insert($$); }
expr: POWER_TOK '(' expr ',' expr ')'    { $$ = new SymbInt(funcall("power", *$3, *$5));         parse.i_heap.insert($$); }
expr: '(' expr ')'                       { $$ = $2; } // reuse $2


APs: /* empty */
  |  AP_TOK AP_decls

AP_decls: AP_decl
  |       AP_decls AP_decl

AP_decl: ID_TOK ';' { parse.AP_decl(*$1); }

%%


void yyerror(Parse& parse, char const *msg) { parse.remove_verbose(); throw CONTROL::Error(string(msg) + " in line " + PPRINT::widthify(parse.line_number)); }

extern FILE* yyin;

int main(const int argc, const char * const * const argv) {
  Input I;
  I.add_switch  ("-v",       "verbose");
  I.add_switch  ("-s",       "print symbol table");
  I.add_argument("filename", "name of SVISS source file");

  try {
    I.get_command_line(argc, argv); }
  catch (Input::Help) {
    return 0; }

  bool verbose      = ( I.arg_value("-v") == "true" );
  bool print_symtab = ( I.arg_value("-s") == "true" );

  if (verbose)
    I.print_command_line();

  Parse parse(I.arg_value("filename"), I.arg_value("-v"));

  yyin = parse.ifile;
  yyparse(parse);

  if (print_symtab)
    parse.print_symtab();

  return 0; }
