
#include <cctype>

#include <algorithm>
#include <istream>

#include "functions.hh"
#include "symbint.hh"


bool SymbInt::is_identifier(const string& s) {
  return
    (! s.empty()) &&
    (isalpha(s[0]) || s[0] == '_') &&
    (std::find_if(s.begin(), s.end(), Is_not_alnum()) == s.end()); } // [a-zA-Z_][a-zA-Z0-9_]*

void SymbInt::check_identifier() const {
//   if (! is_identifier())
//     CONTROL::warning("SymbInt::SymbInt(string): " + svalue + " " +
// 		     "does not look like an identifier; if operations are supported by SymbInt, please build expressions bottom-up");
}

SymbInt::operator string() const {
  if (is_int_literal())
    return PPRINT::widthify(ivalue);
  if (type == SUM)
    return ( ivalue == 0 ? svalue : PPRINT::widthify(ivalue) + " + " + svalue );
  __SAFE_ASSERT__ ( ivalue != 0 && ivalue != 1 ); // should have been reduced
  if (ivalue == -1)
    return "-" + parentheses(svalue);
  else
    return PPRINT::widthify(ivalue) + " * " + parentheses(svalue); }

// "Last-resort" calculation: if all failes, symbolically combine left and right using op.
// The result is by convention a SUM with integral part 0
SymbInt symbolic(const SymbInt& left, const SymbInt& right, const char& op) {
  SymbInt result;
  result.ivalue = 0; // !! we lose the integral part
  result.svalue = SymbInt::parentheses(string(left)) + " " + string(1, op) + " " + SymbInt::parentheses(string(right));
  result.type = SymbInt::SUM;
  return result; }

SymbInt operator+(const SymbInt& left, const SymbInt& right) {
  if (left == 0)
    return right;
  if (right == 0)
    return left;
  if (left.is_int_literal()) {
    if (right.is_int_literal())                                    // L+L: just add
      return SymbInt(left.ivalue + right.ivalue);
    if (right.is_sum()) {                                          // L+S: add L to S's literal part
      SymbInt result = right;
      result.ivalue += left.ivalue;
      return result; }
    if (right.is_prod()) {                                         // L+P: put L in ivalue, convert P to string and put it in svalue
      SymbInt result;
      result.ivalue = left.ivalue;
      result.svalue = string(right);
      result.type = SymbInt::SUM;
      return result; }}
  if ((left.is_sum() || left.is_prod()) && right.is_int_literal()) // S+L, P+L: flip
    return right + left;
  if (left.is_sum() && right.is_sum()) {                           // S+S: add the literal parts and symbolic parts separately
    SymbInt result;
    result.ivalue = left.ivalue + right.ivalue;
    result.svalue = left.svalue + " + " + right.svalue;
    result.type   = SymbInt::SUM;
    return result; }
  __SAFE_ASSERT__ (left.is_prod() || right.is_prod());             // S+P, P+S, P+P
  __SAFE_ASSERT__ ( (! left.is_int_literal()) && (! right.is_int_literal()));
  return symbolic(left, right, '+'); }

SymbInt operator*(const SymbInt& left, const SymbInt& right) {
  if (left == 0 || right == 0)
    return 0;
  if (left == 1)
    return right;
  if (right == 1)
    return left;
  if (left == -1)
    return right.negation();
  if (right == -1)
    return left.negation();
  if (left.is_int_literal()) {
    if (right.is_int_literal())                                    // L*L
      return SymbInt(left.ivalue * right.ivalue);
    if (right.is_sum()) {                                          // L*S
      SymbInt result;
      result.ivalue = left.ivalue;
      result.svalue = string(right);
      result.type = SymbInt::PROD;
      return result; }
    if (right.is_prod()) {                                         // L*P
      SymbInt result = right;
      result.ivalue *= left.ivalue;
      return result; }}
  if ((left.is_sum() || left.is_prod()) && right.is_int_literal()) // S*L, P*L
    return right * left;
  if (left.is_prod() && right.is_prod()) {                         // P*P
    SymbInt result;
    result.ivalue = left.ivalue * right.ivalue;
    result.svalue = SymbInt::parentheses(left.svalue) + " * " + SymbInt::parentheses(right.svalue);
    result.type   = SymbInt::PROD;
    return result; }
  __SAFE_ASSERT__ (left.is_sum() || right.is_sum());               // S*S, S*P, P*S
  __SAFE_ASSERT__ ( (! left.is_int_literal()) && (! right.is_int_literal()));
  return symbolic(left, right, '*'); }

SymbInt operator/(const SymbInt& left, const SymbInt& right) {
  if (right == 0) {
    cerr << "operator<(SymbInt&, SymbInt&): error: division by 0" << endl; exit (1); }
  if (left.is_int_literal() && right.is_int_literal())
    return SymbInt(left.ivalue / right.ivalue);
  if (left == 0)
    return 0;
  if (right == 1)
    return left;
  if (right == -1)
    return left.negation();
  return symbolic(left, right, '/'); }

SymbInt SymbInt::negation() const {
  if (is_int_literal())
    return SymbInt(-1 * ivalue);
  if (is_sum()) {
    SymbInt result = *this;
    result.ivalue = -1 * result.ivalue;
    result.svalue = "-" + parentheses(result.svalue);
    return result; }
  __SAFE_ASSERT__ (is_prod());
  SymbInt result;
  result.ivalue = 0;
  result.svalue = "-" + parentheses(string(*this));
  result.type   = SymbInt::SUM;
  return result; }

SymbInt funcall(const string& funcname, const SymbInt& x) {
  if (x.is_int_literal()) {
//     cint ix = int(x);
//     if      (funcname == "function_name")
//       return function_name(ix);
  }
  return funcname + "(" + string(x) + ")"; }

SymbInt funcall(const string& funcname, const SymbInt& x, const SymbInt& y) {
  if (x.is_int_literal() && y.is_int_literal()) {
    cint ix = int(x), iy = int(y);
    if      (funcname == "power")
      return LOGARITHM::power(ix, iy);
    else if (funcname == "floor_log")
      return LOGARITHM::floor_log(ix, iy);
    else if (funcname == "ceil_log")
      return LOGARITHM::ceil_log(ix, iy);
  }
  return funcname + "(" + string(x) + ", " + string(y) + ")"; }


std::istream& operator>>(std::istream& in, SymbInt& symbint) {
  string s;
  in >> s;
  if (SymbInt::is_int_literal(s))
    symbint = atoi(s.c_str());
  else
    symbint = s;
  return in; }
