
#ifndef __SYMBINT_HH__
#define __SYMBINT_HH__

#include <cctype>
#include <algorithm>
#include <istream>
#include <ostream>

using std::string;

// Symbolic integer arithmetic. Conventions:
// 1. type is only relevant if ! svalue.empty()
// 2. if type == PROD, ivalue must be != 0 (reduce result to 0), != 1 (reduce result to the SUM 0+svalue).
// 3. Identifiers A are represented as 0+A, not as 1*A. Likewise, -A is represented as 0-A, not as -1 * A.
//    Likewise, 1/A is represented as 0+(1/A) (SUM)
class SymbInt {
private:
  int ivalue;
  string svalue;
  typedef enum { SUM, PROD } Type;
  Type type;

  struct Is_not_alnum { bool operator() (const char& c) { return (! isalnum(c)) && c != '_'; } };
  struct Is_not_digit { bool operator() (const char& c) { return ! isdigit(c); } };

  static inline bool   is_pos_literal(const string& s) { return (! s.empty()) && (std::find_if(s.begin(), s.end(), Is_not_digit()) == s.end()); } //   [0-9]+
  static inline bool   is_int_literal(const string& s) { return is_pos_literal( s[0] == '-' ? s.substr(1) : s ); }                                // -?[0-9]+
  static        bool   is_identifier (const string& s);

public:
  static inline string parentheses(const string& s) { return ( is_int_literal(s) || is_identifier(s) ? s : "(" + s + ")" ); } // optimization

private:
  void check_identifier() const;

public:
  inline bool is_int_literal() const { return svalue.empty(); }
  inline bool is_pos_literal() const { return is_int_literal()      && ivalue >= 0; }
  inline bool is_identifier () const { return is_identifier(svalue) && ivalue == 0; }

private:
  inline bool is_sum () const { return (! is_int_literal()) && type == SUM; }
  inline bool is_prod() const { return (! is_int_literal()) && type == PROD; }

public:
  // conversion. Careful: just as in int(3.5) == 3, conversion to int loses precision: svalue part is cut off!!
  inline operator int   () const { return ivalue; }
         operator string() const;

  // construction. Note to user: to build "A+3", don't pass string "A+3". Instead, use SymbInt("A") + SymbInt(3)
  inline SymbInt() {} // no explicit initializations
  inline SymbInt(const int& i):        ivalue(i), svalue("") {}
  inline SymbInt(const string& value): ivalue(0), svalue(value), type(SUM) { check_identifier(); }
  inline SymbInt(const char* value):   ivalue(0), svalue(value), type(SUM) { check_identifier(); } // for C-style strings

  SymbInt negation () const; //  -X

  // friends: arithmetic
  friend SymbInt operator+ (const SymbInt& left, const SymbInt& right);
  friend SymbInt operator- (const SymbInt& left, const SymbInt& right);
  friend SymbInt operator* (const SymbInt& left, const SymbInt& right);
  friend SymbInt operator/ (const SymbInt& left, const SymbInt& right);

private:
  // "Last-resort" calculation: if all fails, symbolically combine left and right
  // using op. The result is by convention a SUM with integral part 0
  friend SymbInt symbolic(const SymbInt& left, const SymbInt& right, const char& op);

public:
  // comparison
  friend bool operator==(const SymbInt& left, const int& i);
  friend bool operator==(const int& i, const SymbInt& left);

  // stream operators
  friend std::istream& operator>>(std::istream& in,        SymbInt& symbint);
  friend std::ostream& operator<<(std::ostream& out, const SymbInt& symbint); };

SymbInt funcall(const string& funcname, const SymbInt& x);                   // unary
SymbInt funcall(const string& funcname, const SymbInt& x, const SymbInt& y); // binary

inline SymbInt operator-(const SymbInt& left, const SymbInt& right) { return left + right. negation(); }

inline bool operator==(const SymbInt& si, const int& i) { return si.is_int_literal() && int(si) == i; }
inline bool operator==(const int& i, const SymbInt& si) { return si == i; }

inline std::ostream& operator<<(std::ostream& out, const SymbInt& symbint) { out << string(symbint); return out; }

#endif
