
#ifndef __SYMTAB_HH__
#define __SYMTAB_HH__

#include <utility>
#include <algorithm>
#include <list>
#include <iostream>

#include "functions.hh"
#include "triple.hh"

using std::list;

// General symbol table class
template <class E>
class Symtab {
public:
  class EntryRecord: private Triple<string,ushort,E> { // private derivation to make first, second, third private
  public:
    inline EntryRecord(const string& symname, cushort& scope, const E& entry): Triple<string,ushort,E>(symname, scope, entry) {}
    inline const string& symname() const { return this->first;  }    // without "this->", compiler complains??
    inline const ushort& scope  () const { return this->second; }
    inline const E     & entry  () const { return this->third;  }
    inline       E     & entry  ()       { return this->third;  } }; // if we want to change the entry

  typedef typename list<EntryRecord>::const_iterator const_iterator;
  typedef typename list<EntryRecord>::      iterator       iterator;

  // list wrapper. Only iterator access is public
  class Contents {
  private:
    list<EntryRecord> l;
    inline void push_back(const EntryRecord& er) { l.push_back(er); }
  public:
    inline const_iterator begin() const { return l.begin(); }
    inline const_iterator   end() const { return l.  end(); }
    inline       iterator begin()       { return l.begin(); }
    inline       iterator   end()       { return l.  end(); }

    friend class Symtab<E>; }; // seems it is not possible to declare only Symtab<E>::push_back a friend of Contents
                               // Only classes and external functions can be friends?

  Contents contents;

private:
  size_t width;
  void update_width(const string& symname) { width = std::max(width, symname.size()); }

  // for find_if:
  struct EntryRecord_compare {
    const string symname;
    const ushort scope;
    inline EntryRecord_compare(const string& symname, cushort& scope): symname(symname), scope(scope) {}
    inline bool operator() (const EntryRecord& e) const { return e.symname() == symname && e.scope() == scope; } };

  inline const_iterator find(const string& symname, cushort& scope) const {
    return std::find_if(contents.begin(), contents.end(), EntryRecord_compare(symname, scope)); }
  inline iterator find(const string& symname, cushort& scope) {
    return std::find_if(contents.begin(), contents.end(), EntryRecord_compare(symname, scope)); }

public:
  Symtab(): width(0) {}

  inline size_t the_width() const { return width; }
  inline void push_back(const string& symname, cushort& scope, const E& entry) { contents.push_back(EntryRecord(symname, scope, entry)); update_width(symname); }
  inline void push_back(const string& symname,                 const E& entry) { push_back(symname, 0, entry); update_width(symname); } // scope defaults to 0, see next
         void insert   (const string& symname, cushort& scope, const E& entry);
  inline void insert   (const string& symname,                 const E& entry) { insert   (symname, 0, entry); update_width(symname); } // scope defaults to 0, see next
         void remove   (const string& symname, cushort& scope = 0);
  const E& lookup(const string& symname, cushort& scope = 0) const;

  void print(cushort& indent = 0, ostream& out = cout) const;

  // exceptions
  struct ReInsert { string symname; cushort scope; ReInsert(const string& symname, cushort& scope); };
  struct NoRemove { string symname; cushort scope; NoRemove(const string& symname, cushort& scope); };
  struct NotFound { string symname; cushort scope; NotFound(const string& symname, cushort& scope); }; };

template <class E>
void Symtab<E>::insert(const string& symname, cushort& scope, const E& entry) {
  const_iterator record = find(symname, scope);
  if (record != contents.end())
    throw ReInsert(symname, scope);
  push_back(symname, scope, entry);
  update_width(symname); }

template <class E>
void Symtab<E>::remove(const string& symname, const cushort& scope) {
  iterator record = find(symname, scope);
  if (record == contents.end())
    throw NoRemove(symname, scope);
  contents.erase(record); }

template <class E>
const E& Symtab<E>::lookup(const string& symname, cushort& scope) const {
  const_iterator record = find(symname, scope);
  if (record == contents.end())
    throw NotFound(symname, scope);
  return record->entry(); }

template <class E>
void Symtab<E>::print(cushort& indent, ostream& out) const {
  for (const_iterator record = contents.begin(); record != contents.end(); ++record) {
    out << string(indent, ' ') << PPRINT::widthify(record->symname(), width, PPRINT::LEFTJUST)
	<< " (scope " << record->scope() << "): ";
    record->entry()->print(indent, out); out << endl; }}

template<class E>
Symtab<E>::ReInsert::ReInsert(const string& symname, cushort& scope): symname(symname), scope(scope) {
  cerr << "symbol table error: symbol " << symname << " (scope " << scope << ") already exists" << endl; }

template<class E>
Symtab<E>::NoRemove::NoRemove(const string& symname, cushort& scope): symname(symname), scope(scope) {
  cerr << "symbol table error: cannot remove symbol " << symname << " (scope " << scope << ") since it does not exist" << endl; }

template<class E>
Symtab<E>::NotFound::NotFound(const string& symname, cushort& scope): symname(symname), scope(scope) {
  cerr << "symbol table error: symbol " << symname << " (scope " << scope << ") not found" << endl; }

#endif
