#include <stdlib.h>

#include <iostream>
#include <iomanip>
#include <algorithm>
#include <assert.h>
#include <string>
#include <vector>

#include "cuddObj.hh"
#include "utool.hh"
#include "functions.hh"

extern UCudd* ucudd;

char bit2char(cushort& bit) {
  switch (bit) {
  case 0:  return '0';
  case 1:  return '1';
  case 2:  return '-';
  default: throw CONTROL::Error("bit2char: 0 <= bit <= 2 required"); }}


//////////////////////////////////
// Definition of Class UCudd
//////////////////////////////////

void UCudd::createStateVars(cushort& size, cushort& start, cushort& step) {
  currStateVars = new int[size];
  nextStateVars = new int[size];
  currStateVector.reserve(size);
  nextStateVector.reserve(size);
  for (ushort i = 0, bit = start; i < size; ++i, bit += step) {
    currStateVars[i] = bit;
    nextStateVars[i] = bit + 1;
    currStateVector.push_back(bit);
    nextStateVector.push_back(bit + 1); }
  currStateCube = IndicesToCube(currStateVars, size);
  nextStateCube = IndicesToCube(nextStateVars, size); }

void UCudd::deleteStateVars() { nextStateCube = currStateCube = UBDD::Zero(); nextStateVector.clear(); currStateVector.clear(); delete nextStateVars; delete currStateVars; }

////////////////////////////////////////////////////////////
///////////// For printing routines ////////////////////////
////////////////////////////////////////////////////////////

void UCudd::print_all(const int* const bits, ostream& out) const {
  for (ushort i = 0; i < ReadSize(); ++i)
    out << bit2char(bits[i]); }

void UCudd::print_curr_state(const int* const bits, ostream& out) const {
  for (ushort i = 0; i < currStateVector.size(); ++i)
    out << bit2char(bits[ucudd->currStateVars[i]]); }

void UCudd::print_next_state(const int* const bits, ostream& out) const {
  for (ushort i = 0; i < currStateVector.size(); ++i)
    out << bit2char(bits[ucudd->nextStateVars[i]]); }


////////////////////////////////////
// Definition of Class UBDD
////////////////////////////////////

// Rewrite *this, in terms of variables from, from + step, ..., into to, to + step, ...
UBDD UBDD::shift (cushort& from, cushort& to, cushort& size, cushort& step) const {
  UBDD result = *this;
  for (ushort i = 0, from_bit = from, to_bit = to; i < size; ++i, from_bit += step, to_bit += step)
    result = result.Compose(ucudd->bddVar(to_bit), from_bit);
  return result; }

UBDD UBDD::inverse() const {
  // cout << "calling swap ... "; cout.flush(); // debug
  cushort size = ucudd->currStateVector.size();
  DdNode** xp = ALLOC(DdNode*, size);
  DdNode** xq = ALLOC(DdNode*, size);
  for (ushort b = 0; b < size; ++b) {
    xp[b] = ucudd->bddVar(ucudd->currStateVars[b]).getNode();
    xq[b] = ucudd->bddVar(ucudd->nextStateVars[b]).getNode(); }

  DdNode* swapped = Cudd_bddSwapVariables(ucudd->getManager(), getNode(), xp, xq, size); // done in C, for efficiency

  FREE(xp); FREE(xq);
  ucudd->checkReturnValue(swapped);
  // cout << "done" << endl; // debug
  return UBDD(ucudd, swapped); }

UBDD UBDD::invariant(const vector<ushort>& bits) {
  UBDD invariant = One();
  for (int i = bits.size() - 1; i >= 0; --i) {
    cushort& b = bits[i];
    invariant &= equal(b, b+1); }
  return invariant; }

UBDD UBDD::square() const {
  UBDD Rxy = *this;
  // obtain Ryz using these steps: Rxy --(invert x,y)--> Ryx --(shift x to z)--> Ryz
  UBDD Ryx = Rxy.inverse();
  UBDD Ryz = Ryx.shift(0, 2, ucudd->currStateVector.size(), 3);
  UBDD Rxyz = Rxy & Ryz;
  UBDD Rxz = Rxyz.ExistAbstract(ucudd->nextStateCube);
  UBDD result = Rxz.shift(2, 1, ucudd->currStateVector.size(), 3);
  return result; }

UBDD UBDD::aggregate(const vector<UBDD>& args, cushort& start, const vector<UBDD>::size_type& size, Associative& assoc) {
  const vector<UBDD>::size_type half = size / 2;
  //cout << "aggregate(args, start = " << start << ", size = " << size << "); total number of args is " << args.size() << endl;
  return ( size == 1 ? args[start] : assoc.op(aggregate(args, start, half, assoc), aggregate(args, start + half, size - half, assoc)) ); }

UBDD UBDD::cubify(const vector<ushort>& bits) {
  UBDD cube = UBDD::One();
  for (ushort b = 0; b < bits.size(); ++b)
    cube &= ucudd->bddVar(bits[b]);
  return cube; }

int* UBDD::UBDD2Cube() const {
  int** cube = new int*;
  CUDD_VALUE_TYPE* value = new CUDD_VALUE_TYPE;
  assert (FirstCube(cube, value));
  int* result = *cube;
  delete value;
  delete cube;
  return result; }

// Takes a vector of UBDDs and writes them at variables in bits (conversion functional to relational representation)
// This function is slower when you fill result from right to left!
UBDD UBDD::funs2rel(const vector<UBDD>& f, const vector<ushort>& bits) {
  __SAFE_ASSERT__ (f.size() == bits.size());
  UBDD result = One();
  ushort b = 0;
  for (vector<UBDD>::const_iterator fi = f.begin(); fi != f.end(); ++fi, ++b)
    result &= ucudd->bddVar(bits[b]).Eqv(*fi);
  return result; }

// This is more cumbersome than equal(bits, l), since I first must build the number from right to left,
// then build then result BDD from left to right (highest-order bit first!)
UBDD UBDD::greater(const vector<ushort>& bits, culong& l) {
  cushort& numbits = bits.size();
  __SAFE_ASSERT__ (numbits >= std::max(LOGARITHM::ceil_log(l + 1, 2), ushort(1)));
  vector<bool> binary(numbits, false); // by default, digits are 0. This takes care of leading zeros
  ulong r = l;
  int b = numbits - 1; // last bit; fill number from right to left
  while (b >= 0) {
    binary[b] = ( r % 2 != 0 );
    r /= 2;
    --b; }
  __SAFE_ASSERT__ (r == 0);
  UBDD greater = Zero();
  UBDD same   = One();
  for (ushort b = 0; b < numbits; ++b) {
    UBDD curr = ucudd->bddVar(bits[b]);
    if (! binary[b]) { // only if l has a 0 at position b, there is a chance for bits to be greater than l
      greater |= same & curr;
      same   &= ! curr; }
    else
      same   &=   curr; }
  return greater; }

UBDD UBDD::equal(const vector<ushort>& bits, culong& l) {
  cushort& numbits = bits.size();
  __SAFE_ASSERT__ (numbits >= std::max(LOGARITHM::ceil_log(l + 1, 2), ushort(1))); // we need ceil_log(l + 1, 2) bits to store the number l, except when l is 0, where we need 1 bit
  UBDD equal = One();
  ulong r = l;
  int b = numbits - 1; // last bit; fill number from right to left
  while (b >= 0) {
    UBDD curr = ucudd->bddVar(bits[b]);
    equal &= ( r % 2 == 0 ? ! curr : curr );
    r /= 2;
    --b; }
  __SAFE_ASSERT__ (r == 0);
  return equal; }

UBDD UBDD::greater(const vector<ushort>& x, const vector<ushort>& y) {
  __SAFE_ASSERT__ (x.size() == y.size());
  UBDD greater = Zero();
  UBDD same   = One();
  for (ushort b = 0; b < x.size(); ++b) {
    UBDD curr_1 = ucudd->bddVar(x[b]);
    UBDD curr_2 = ucudd->bddVar(y[b]);
    greater |= same & curr_1 & (! curr_2); // v1 > v2: v1 == 1, v2 == 0
    same    &= curr_1.Eqv(curr_2); }
  return greater; }

// Could also implement this as ! greater(x, y) & ! less(x, y)
UBDD UBDD::equal(const vector<ushort>& x, const vector<ushort>& y) {
  __SAFE_ASSERT__ (x.size() == y.size());
  UBDD equal = One();
  for (ushort b = 0; b < x.size(); ++b) {
    UBDD curr_1 = ucudd->bddVar(x[b]);
    UBDD curr_2 = ucudd->bddVar(y[b]);
    equal &= curr_1.Eqv(curr_2); }
  return equal; }

UBDD UBDD::dec(const vector<ushort>& source, const vector<ushort>& target, UBDD& overflow) {
  __SAFE_ASSERT__ (source.size() == target.size());
  UBDD sub = One();
  UBDD dec = One();
  for (int b = source.size() - 1; b >= 0; --b) {
    UBDD x = ucudd->bddVar(source[b]);
    UBDD X = x.Xor(sub);
    dec &= ucudd->bddVar(target[b]).Eqv(X);
    sub &= ! x; }
  overflow = sub;
  return dec; }

UBDD UBDD::inc(const vector<ushort>& source, const vector<ushort>& target, UBDD& overflow) {
  __SAFE_ASSERT__ (source.size() == target.size());
  UBDD add = One();
  UBDD inc = One();
  for (int b = source.size() - 1; b >= 0; --b) {
    UBDD x = ucudd->bddVar(source[b]);
    UBDD X = x.Xor(add);
    inc &= ucudd->bddVar(target[b]).Eqv(X);
    add &= x; }
  overflow = add;
  return inc; }

// If *this is Zero, abort. Otherwise, return a "random" cube in the on-set of *this. part determines which bits to consider.
// max_dont_care determines how many don't-cares (-) are allowed in the cube. If 0 (the default), the result is really a single cube.
UBDD UBDD::randomCube(cushort& max_dont_care, const UCudd::Part& part) const {
  UBDD Z = *this;
  if (Z.empty())
    throw CONTROL::Error("UBDD::randomCube: given BDD is empty");
  UBDD randomCube = UBDD::One();
  cushort start = ( part == UCudd::next ? 1 : 0 ); // If "next", start at 1,         otherwise ("curr" or "all")  at 0
  cushort step  = ( part == UCudd::all  ? 1 : 2 ); // If "all",  consider every bit, otherwise ("curr" or "next") every second
  ushort remaining_dont_care = max_dont_care;
  Random::set_seed();
  for (ushort b = start; b < ucudd->ReadSize(); b += step) {
    UBDD var = ucudd->bddVar(b);
    UBDD neg = Z & !var;
    UBDD pos = Z &  var;
    __SAFE_ASSERT__ ( !(neg.empty()) || !pos.empty());
    if      (neg.empty())   // we have no choice
      randomCube &=  var;
    else if (pos.empty())   // we have no choice
      randomCube &= !var;
    else {
      ushort c = Random::integer(0, ( remaining_dont_care == 0 ? 1 : 2 ));
      switch (c) {
      case 0: randomCube &=  var;    break;
      case 1: randomCube &= !var;    break;
      case 2: --remaining_dont_care; break; // randomCube is not changed in this case - var remains don't care
      default: throw CONTROL::Error("UBDD::randomCube: 0 <= c <= 2 required"); }}
    Z &= randomCube; }                  // for efficiency, fix the value of var in Z
  return randomCube; }

// Choose m random elements over the bits as indicated by part
// Think about a function to return m elements from a given BDD, instead of from ONE.
// Can it be done recursively? Say input is (m, source), first pick a
// (small?) random tuple t from source that has at most m members. Then
// keep those and call the function on (m-|t|, source \ t).
UBDD UBDD::random(culong& m, const UCudd::Part& part) {
  UBDD result = UBDD::Zero();
  ulong remaining = m;
  while (remaining > 0) {
    UBDD t = (!result).randomCube(LOGARITHM::floor_log(remaining, 2), part);
    result |= t;
    ulong size = result.cardinality(part);
    remaining = m - size; }
  return result; }

ushort UBDD::lines_per_page = 35;

// Enumerates all input bit combinations that lead to output 1, and processes them using the read-only function passed
void UBDD::enumerate(Read_bits& read_bits, cushort& lines_per_page, ostream& out) const {
  if (ucudd->ReadSize() == 0) {
    CONTROL::warning("UBDD::enumerate: no variables declared");
    return; }
  if (empty()) // if empty, the loop below does nothing
    out << "<empty>" << endl;
  else {
    DdGen *gen; int* cube; CUDD_VALUE_TYPE value;
    if (lines_per_page > 0) { // this test factored out for efficiency
      ushort line_number = 0;
      Cudd_ForeachCube(ucudd->getManager(), getNode(), gen, cube, value) { // no cuddObj version of this function
	read_bits(cube);
	if (++line_number == lines_per_page) {
	  if (CONTROL::wait_yes_no("Continue?", out))
	    line_number = 0;
	  else
	    return; }}}
    else {
      Cudd_ForeachCube(ucudd->getManager(), getNode(), gen, cube, value)
	read_bits(cube); }}}

// Search Routines (existential only) vs. Fixpoint Routines.
// Search Routines apply the "tau" operator (usually an image operator)
// only to the unexplored states, whereas Fixpoint Routines apply them to
// all states collected so far. If the operator is monotone, this results
// in the same fixpoint. "Frontier Set optimization"
// There can be, however, differences in sizes of intermediate BDDs, simply
// because the intermediate sets are different. Therefore, it still makes
// sense to consider both methods.

// Fixpoint (greatest or least, depending on tau): [mu|nu]Z.tau(Z).
UBDD UBDD::FP(Tau& tau, const string& id, const bool& verbose) const {
  UBDD Z = *this, Zp;
  Report r(id, verbose);
  do {
    Zp = Z;
    Z = tau(Z);
    r.print(Z); }
  while (Z != Zp);
  return Z; }

// Search from *this, using image operation supplied by tau
UBDD UBDD::Search(Tau& tau, const string& id, const bool& verbose) const {
  UBDD Unexplored = *this, Reached = Unexplored;
  Report r(id, verbose);
  while (! Unexplored.empty()) {
    UBDD Successors;
    Successors = tau(Unexplored);
    Unexplored = Successors & (! Reached); // if used in the quotient, ! should really be neg. But it is OK here, since Successors \subset Rep
    Reached |= Unexplored;
    r.print(Unexplored, Reached); }
  return Reached; }

UBDD UBDD::iterSquare  (const bool& verbose) const {   IterSquareStep tau(*this); return UBDD::identity() | FP(tau, "IterSquare",   verbose); }
UBDD UBDD::transClosure(const bool& verbose) const { TransClosureStep tau(*this); return UBDD::identity() | FP(tau, "TransClosure", verbose); }

// Spit out some statistics during fixpoint computation
void Report::print(const UBDD& iterate, ostream& out) {
  using namespace PPRINT;
  if (verbose) {
    ++l;
    ulong delta_time = util_cpu_time() - ucudd->global_time;
    out << "step " << id       << ": " << widthify(l,                              3, RIGHTJUST) << ", "
	<< "iterate Z"         << ": " << widthify(iterate.nodeCount(),            9, RIGHTJUST) << " nodes, "
	<< "peakLiveNodeCount" << ": " << widthify(ucudd->ReadPeakLiveNodeCount(), 9, RIGHTJUST) << ", "
	<< "sec since start"   << ": " << widthify(delta_time / 1000,              5, RIGHTJUST) << endl; }}

void Report::print(const UBDD& Unexplored, const UBDD& Reached, ostream& out) {
  using namespace PPRINT;
  if (verbose) {
    ++l;
    ulong delta_time = util_cpu_time() - ucudd->global_time;
    out << "step " << id       << ": " << widthify(l,                                         3, RIGHTJUST) << ", "
	<< "Unexplored"        << ": " << widthify(Unexplored.nodeCount(),                    9, RIGHTJUST) << " nodes, "
	<< "Reached"           << ": " << widthify(Reached.nodeCount(),                       9, RIGHTJUST) << " nodes, "
	<< "peakLiveNodeCount" << ": " << widthify(ucudd->ReadPeakLiveNodeCount(),            9, RIGHTJUST) << ", "
	<< "sec since start"   << ": " << widthify(delta_time / 1000,              5, RIGHTJUST) << endl; }}

culong UBDD::Cardinality_Overflow::Max = ulong(0) - 1; // hack: MAX_ULONG

// Returns the number of elements in the set represented by *this,
// considered over the state variables with the given indices.
ulong UBDD::cardinality(const UCudd::Part& part) const {
  const int* const indices = ( part == UCudd::curr ? ucudd->currStateVars : ucudd->nextStateVars );
  ulong result = 0;
  DdGen *gen; int* cube; CUDD_VALUE_TYPE value;
  Cudd_ForeachCube(ucudd->getManager(), getNode(), gen, cube, value) {
    ulong count_per_cube = 1;
    if (part == UCudd::all) {
      for (ushort i = 0; i < ucudd->ReadSize(); ++i)
        if (cube[i] == 2)
	  Cardinality_Overflow::doublE(count_per_cube); } // a don't care position ('-') doubles the count per state
    else
      for (ushort i = 0; i < ucudd->currStateVector.size(); ++i)
        if (cube[indices[i]] == 2)
	  Cardinality_Overflow::doublE(count_per_cube);   // dito
    Cardinality_Overflow::inc(result, count_per_cube); }
  return result; }

