#ifndef __UTOOL__
#define __UTOOL__

#include <string>
#include <iostream>
#include <vector>
#include <algorithm>

#include "functions.hh"
#include "cuddObj.hh"


char bit2char(cushort& bit);


///////////////////////////////////////////////////
// Declaration of Class UCudd
///////////////////////////////////////////////////

class UBDD;

class UCudd: public Cudd {
private:
  int*            currStateVars;   // these have to be int* since they are int* in cudd (can't cast to short*)
  int*            nextStateVars;
  vector<ushort>  currStateVector;
  vector<ushort>  nextStateVector;
  BDD             currStateCube;   // must use BDD here since UBDD is still undefined
  BDD             nextStateCube;

  void createStateVars(cushort& size, cushort& start = 0, cushort& step = 2);
  void deleteStateVars();

public:
  ulong global_time;

  enum Part { curr, next, all };

  inline  UCudd(cushort& numberOfVars = 0): Cudd(numberOfVars, 0, CUDD_UNIQUE_SLOTS, CUDD_CACHE_SLOTS, 0), global_time(util_cpu_time()) { createStateVars(numberOfVars / 2); }
  inline ~UCudd() { deleteStateVars(); }

  inline void updateStateVars(cushort& size, cushort& start = 0, cushort& step = 2) { deleteStateVars(); createStateVars(size, start, step); }

  UBDD bddVar(cushort& i);
  UBDD newVar(cushort& i);

  inline const vector<ushort>& currStates() const { return currStateVector; }
  inline const vector<ushort>& nextStates() const { return nextStateVector; }

  inline uint ReadPeakLiveNodeCount() { return Cudd_ReadPeakLiveNodeCount(getManager()); } // function missing in cuddObj.hh

  void print_all       (const int* const bits, ostream& out = cout) const;
  void print_curr_state(const int* const bits, ostream& out = cout) const;
  void print_next_state(const int* const bits, ostream& out = cout) const;

  friend class UBDD; };


extern UCudd* ucudd;


////////////////////////////////////////////////////////////
// Definition of printing routines /////////////////////////
////////////////////////////////////////////////////////////

// Class with a read-only operator to access bits
struct Read_bits { virtual void operator ()(const int* const bits) = 0; virtual ~Read_bits() {} };

// All bits
struct Print_bits: public Read_bits {
  ostream& out;
  inline Print_bits(ostream& out = cout): out(out) {}
  inline void operator()(const int* const bits) { out << "  "; ucudd->print_all(bits, out); out << endl; } };

// Only current state bits
struct Print_curr_state: public Print_bits {
  inline Print_curr_state(ostream& out = cout): Print_bits(out) {}
  inline void operator()(const int* const bits) { out << "  "; ucudd->print_curr_state(bits, out); out << endl; } };

// Only next state bits
struct Print_next_state: public Print_bits {
  inline Print_next_state(ostream& out = cout): Print_bits(out) {}
  inline void operator()(const int* const bits) { out << "  "; ucudd->print_next_state(bits, out); out << endl; } };

// Curr bits, then next bits (instead of interleaved)
struct Print_curr_next: public Print_bits {
  inline Print_curr_next(ostream& out = cout): Print_bits(out) {}
  inline void operator()(const int* const bits) { out << "  "; ucudd->print_curr_state(bits, out); out << "  ==>  "; ucudd->print_next_state(bits, out); out << endl; } };

// Generic image operation
struct Image;

struct EX_Image;
struct AX_Image;
struct EY_Image;

// General tau. All we know is that we must provide a tau operator. It is assumed to take a BDD and output a new one.
// The operator is non-const since it may have to update some class data
struct Tau { virtual UBDD operator ()(const UBDD& Z) = 0; virtual ~Tau() {} };


////////////////////////////////////
// Declaration of Class UBDD
////////////////////////////////////

class UBDD: public BDD {
public:

  // Constructors
  inline UBDD()                                  : BDD()                    {} // empty UBDD
  inline UBDD(const UBDD& from)                  : BDD(from)                {} // copy constructor
  inline UBDD(const  BDD& from)                  : BDD(from)                {} // conversion BDD->UBDD
  inline UBDD(UCudd* bddManager, DdNode* bddNode): BDD(bddManager, bddNode) {}

  // Constants
  static inline UBDD Zero() { return ucudd->bddZero(); } static inline UBDD False() { return Zero(); }
  static inline UBDD One () { return ucudd->bddOne (); } static inline UBDD True () { return One (); }

  // Boolean operations. These must be redefined since they are supposed to return a UBDD, not a BDD.
  // Further, they must be members, since BDD defines them as members (ISO complains if I define them as friends).
  inline UBDD operator &(const UBDD& other) const { return BDD::operator&(other); }
  inline UBDD operator |(const UBDD& other) const { return BDD::operator|(other); }
  inline UBDD Eqv       (const UBDD& other) const { return Xnor          (other); } // alternate name
  inline UBDD Exor      (const UBDD& other) const { return Xor           (other); } // alternate name
  inline UBDD Diff      (const UBDD& other) const { return ((*this) & ! other) | (other & ! (*this)); } // symmetric difference
  inline UBDD operator !() const { return BDD::operator!(); }

  inline UBDD IfThen(const UBDD& Then) const { return (! *this) | Then; }

  inline bool zero() const { return (*this) == Zero(); } inline bool empty() const { return zero(); }
  inline bool one () const { return (*this) == One (); } inline bool full () const { return  one(); }
  inline bool subset(const UBDD& B) const { return ( (*this) & (! B) ) == UBDD::Zero(); }

  // Rewrite *this, in terms of variables from, from + step, ..., into to, to + step, ...
  UBDD shift(cushort& from, cushort& to, cushort& size, cushort& step) const;
  inline UBDD shift_to_curr  () const { return shift(ucudd->nextStateVector[0], ucudd->currStateVector[0], ucudd->currStateVector.size(), ucudd->currStateVector[1] - ucudd->currStateVector[0]); }
  inline UBDD shift_to_next  () const { return shift(ucudd->currStateVector[0], ucudd->nextStateVector[0], ucudd->currStateVector.size(), ucudd->currStateVector[1] - ucudd->currStateVector[0]); }

  UBDD inverse() const; // R^{-} by swapping current and next state vars (R = *this)
  static UBDD invariant(const vector<ushort>& bits); // <bits>  == <BITS>
  static inline UBDD identity() { return invariant(ucudd->currStateVector); } // all curr state variables equal their respective next state
  static inline UBDD equal(cushort& i, cushort& j)    { return ucudd->bddVar(i).Eqv(ucudd->bddVar(j)); } // var i = var j

  UBDD square() const; // Assumption: R is written in terms of 1st & 2nd copy of state variables
  UBDD iterSquare  (const bool& verbose = false) const;
  UBDD transClosure(const bool& verbose = false) const;

  struct Associative;
  static UBDD aggregate(const vector<UBDD>& args, cushort& start, const vector<UBDD>::size_type& size, Associative& assoc);
  static UBDD aggregate(const vector<UBDD>& args, Associative& assoc);

  // some examples of associative operations:
  struct Conjunction;
  struct Disjunction;
  static UBDD conjoin(const vector<UBDD>& conjuncts);
  static UBDD disjoin(const vector<UBDD>& disjuncts);

  static        UBDD cubify(const vector<ushort>& bits);
  static inline UBDD bit   (const vector<ushort>& bits) { __SAFE_ASSERT__ (bits.size() == 1); return ucudd->bddVar(bits[0]); } // special case: Boolean variable
  int* UBDD2Cube() const;

  inline UBDD existsCurrState() const { return ExistAbstract(ucudd->currStateCube); } // { y: \exists x: (x,y) \in *this}
  inline UBDD existsNextState() const { return ExistAbstract(ucudd->nextStateCube); } // { x: \exists y: (x,y) \in *this}
  inline UBDD forallCurrState() const { return  UnivAbstract(ucudd->currStateCube); } // { y: \forall x: (x,y) \in *this}
  inline UBDD forallNextState() const { return  UnivAbstract(ucudd->nextStateCube); } // { x: \forall y: (x,y) \in *this}

  inline UBDD existsBits(const vector<ushort>& bits) { return ExistAbstract(cubify(bits)); }
  inline UBDD forallBits(const vector<ushort>& bits) { return  UnivAbstract(cubify(bits)); }

  static UBDD funs2rel(const vector<UBDD>& f, const vector<ushort>& bits);           // forall i: bddVar(bits_i) = f_i

  // Arithmetic. For the first group of 5 functions, if __SAFE_COMPUTATION__ is turned on, we check whether the bit vector bits is long enough
  // to hold l; if not we claim the comparison makes no sense. This can lead to a surprise when using the "less" function: Say you have
  // a variable x with range [0..3]. Then 2 bits will be reserved for x. Thus, the test less(x, 4) is illegal! (4 cannot be written with 2 bits.)
  // Solution: use the equivalent test atMost(x, 3) instead.
         static UBDD greater(const vector<ushort>& bits, culong& l);                       // <bits>  >  l
         static UBDD equal  (const vector<ushort>& bits, culong& l);                       // <bits>  == l
  inline static UBDD less   (const vector<ushort>& bits, culong& l) { return (! equal(bits, l)) & (! greater(bits, l)); }
  inline static UBDD atMost (const vector<ushort>& bits, culong& l) { return ! greater(bits, l); }
  inline static UBDD atLeast(const vector<ushort>& bits, culong& l) { return ! less   (bits, l); }

  // The following 5 functions have the same type in both arguments. Thus, all 5 can be defined using an implementation of "greater".
  // For the previous 5, this is not sufficient since, for example, a < b cannot be expressed as b > a (type error)
         static UBDD greater(const vector<ushort>& x, const vector<ushort>& y); // x > y
         static UBDD equal  (const vector<ushort>& x, const vector<ushort>& y);
  inline static UBDD less   (const vector<ushort>& x, const vector<ushort>& y) { return greater(y, x); }
  inline static UBDD atMost (const vector<ushort>& x, const vector<ushort>& y) { return ! greater(x, y); }
  inline static UBDD atLeast(const vector<ushort>& x, const vector<ushort>& y) { return ! less   (x, y); }

  static inline UBDD equal(const vector<ushort>& x, const UBDD& y) { return bit(x).Eqv(y); } // special case: Boolean variable

  // The following functions perform modulo increment and decrement; incrementing 11111111 results in 00000000
         static UBDD dec(const vector<ushort>& source, const vector<ushort>& target, UBDD& overflow); // target = source - 1"
         static UBDD inc(const vector<ushort>& source, const vector<ushort>& target, UBDD& overflow); // target = source + 1"
  inline static UBDD dec(const vector<ushort>& source, const vector<ushort>& target) { UBDD overflow; return dec(source, target, overflow); }  // dito, except that overflow is not returned
  inline static UBDD inc(const vector<ushort>& source, const vector<ushort>& target) { UBDD overflow; return inc(source, target, overflow); }  // dito, except that overflow is not returned

  // Shortcuts for bit variables
  inline static UBDD False(const vector<ushort>& x) { __SAFE_ASSERT__ (x.size() == 1); return UBDD::equal(x, false); }
  inline static UBDD True (const vector<ushort>& x) { __SAFE_ASSERT__ (x.size() == 1); return UBDD::equal(x, true ); }

  inline bool value(int* inputs) const { return Eval(inputs).getNode()->type.value > 0.5; }  // Value of the boolean function given input vector

  UBDD randomCube(cushort& max_dont_care = 0, const UCudd::Part& part = UCudd::all) const; // random cube contained in *this
  static UBDD random(culong& m, const UCudd::Part& part = UCudd::all);                     // random BDD (i.e. random boolean function)

  // Functions to enumerate all cubes in a BDD, usually for printing (debugging)

  static ushort lines_per_page; // enumerate: how many lines to print before pausing

  void enumerate(Read_bits& read_bits, cushort& lines_per_page = 0, ostream& out = cout) const;
  inline void enumerate           (ostream& out = cout) const { Print_bits       p(out); enumerate(p, lines_per_page, out); }
  inline void enumerate_curr_state(ostream& out = cout) const { Print_curr_state p(out); enumerate(p, lines_per_page, out); }
  inline void enumerate_next_state(ostream& out = cout) const { Print_next_state p(out); enumerate(p, lines_per_page, out); }
  inline void enumerate_curr_next (ostream& out = cout) const { Print_curr_next  p(out); enumerate(p, lines_per_page, out); }

  ulong cardinality(const UCudd::Part& part = UCudd::all) const; // number of elements in the BDD (viewed as a set)
  struct Cardinality_Overflow {                                  // thrown when cardinality too large
    static culong Max;
    static inline void inc   (ulong& x, culong& y) { if (x > Max - y) throw Cardinality_Overflow(); x += y; }
    static inline void doublE(ulong& x)            { if (x > Max / 2) throw Cardinality_Overflow(); x *= 2; } };

  ///////////////////////////////////////////////////////////////
  // Functions related to BDDs representing transition systems //
  ///////////////////////////////////////////////////////////////

  // Standard fixed-point routines. *this is the starting point of the FP computation/search
  UBDD FP    (Tau& tau, const string& id, const bool& verbose = false) const;
  UBDD Search(Tau& tau, const string& id, const bool& verbose = false) const;

  // Image operators: backward (X) and forward (Y)
  inline UBDD EX(const UBDD& Z) const { UBDD Zp = Z.shift_to_next(); return (*this & Zp).existsNextState(); }
  inline UBDD AX(const UBDD& Z) const { UBDD Zp = Z.shift_to_next(); return   IfThen(Zp).forallNextState(); }
  inline UBDD EY(const UBDD& Z) const { UBDD result = (*this & Z).existsCurrState(); return result.shift_to_curr(); }
  inline UBDD AY(const UBDD& Z) const { UBDD result =   IfThen(Z).forallCurrState(); return result.shift_to_curr(); }

  // These should really be implemented using direct product abstractions as below,
  // but they turned out less efficient than the above, especially for EY. I'm not sure why
  // inline UBDD EX(const UBDD& Z) const { UBDD Zp = Z.shift_to_next(); return AndAbstract(Zp, ucudd->nextStateCube); }
  // inline UBDD EY(const UBDD& Z) const { UBDD result = AndAbstract(Z, ucudd->currStateCube); return result.shift_to_curr(); }

  // Alternative names for image operators
  inline UBDD existBackIm(const UBDD& Z) const { return EX(Z); }
  inline UBDD  univBackIm(const UBDD& Z) const { return AX(Z); }
  inline UBDD existForwIm(const UBDD& Z) const { return EY(Z); }
  inline UBDD  univForwIm(const UBDD& Z) const { return AY(Z); }

  // Fixpoint Routines; Existential Search. They use the tau function classes and are therefore defined below
  // Note that the _inf routines use the same tau operators as search, but they perform FP instead of search.
  static UBDD EF    (const UBDD& f,                EX_Image& EX, const bool& verbose = false); // EF f
  static UBDD EG    (const UBDD& f,                EX_Image& EX, const bool& verbose = false); // EG f
  static UBDD EU    (const UBDD& f, const UBDD& g, EX_Image& EX, const bool& verbose = false); // E  f U g (strong 'until')
  static UBDD EP    (const UBDD& f,                EY_Image& EY, const bool& verbose = false); // EP f (past-time EF)
  static UBDD AF    (const UBDD& f,                AX_Image& AX, const bool& verbose = false); // AF f
  static UBDD AG    (const UBDD& f,                AX_Image& AX, const bool& verbose = false); // AG f
  static UBDD AU    (const UBDD& f, const UBDD& g, AX_Image& AX, const bool& verbose = false); // A  f U g (dito)
  static UBDD EX_INF(const UBDD& f,                EX_Image& EX, const bool& verbose = false); // EXEXEX... f (may not terminate)
  static UBDD EY_INF(const UBDD& f,                EY_Image& EY, const bool& verbose = false); // EYEYEY... f (may not terminate)
  static UBDD FS    (const UBDD& f,                EY_Image& EY, const bool& verbose = false); // f --R^{+}-->
  static UBDD BS    (const UBDD& f,                EX_Image& EX, const bool& verbose = false); // f --R^{-}-->

  // The same routines, with a predefined standard image operator
  inline UBDD EF    (const UBDD& f,                const bool& verbose = false) const;
  inline UBDD EG    (const UBDD& f,                const bool& verbose = false) const;
  inline UBDD EU    (const UBDD& f, const UBDD& g, const bool& verbose = false) const;
  inline UBDD EP    (const UBDD& f,                const bool& verbose = false) const;
  inline UBDD AF    (const UBDD& f,                const bool& verbose = false) const;
  inline UBDD AG    (const UBDD& f,                const bool& verbose = false) const;
  inline UBDD AU    (const UBDD& f, const UBDD& g, const bool& verbose = false) const;
  inline UBDD EX_INF(const UBDD& f,                const bool& verbose = false) const;
  inline UBDD EY_INF(const UBDD& f,                const bool& verbose = false) const;
  inline UBDD FS    (const UBDD& f,                const bool& verbose = false) const;
  inline UBDD BS    (const UBDD& f,                const bool& verbose = false) const;
};

struct Image {
  const UBDD R;
  inline Image(const UBDD& R): R(R) {}
  virtual UBDD operator ()(const UBDD& Z) = 0;
  virtual ~Image() {} };

struct EX_Image: public Image { inline EX_Image(const UBDD& R): Image(R) {} inline UBDD operator()(const UBDD& Z) { return R.EX(Z); } };
struct AX_Image: public Image { inline AX_Image(const UBDD& R): Image(R) {} inline UBDD operator()(const UBDD& Z) { return R.AX(Z); } };
struct EY_Image: public Image { inline EY_Image(const UBDD& R): Image(R) {} inline UBDD operator()(const UBDD& Z) { return R.EY(Z); } };

struct IterSquareStep: public Tau {
  UBDD R_i; // is updated to contain R, R^2, R^4, ...
  inline IterSquareStep(const UBDD& R): R_i(R) {}
  inline UBDD operator()(const UBDD& Z) { R_i = R_i.square(); return Z | R_i; } };

struct TransClosureStep: public Tau {
  const UBDD R;
  inline TransClosureStep(const UBDD& R): R(R) {}
  inline UBDD operator()(const UBDD& Z) { return R | Z.square(); } };

// These functions and classes need UBDD to be defined
inline UBDD UCudd::bddVar(cushort& i) { __SAFE_ASSERT__ (i <  ReadSize()); return Cudd::bddVar(i); }
inline UBDD UCudd::newVar(cushort& i) { __SAFE_ASSERT__ (i >= ReadSize()); return Cudd::bddVar(i); }

inline UBDD UBDD::EF    (const UBDD& f,                const bool& verbose) const { EX_Image I(*this); return EF    (f,    I, verbose); }
inline UBDD UBDD::EG    (const UBDD& f,                const bool& verbose) const { EX_Image I(*this); return EG    (f,    I, verbose); }
inline UBDD UBDD::EU    (const UBDD& f, const UBDD& g, const bool& verbose) const { EX_Image I(*this); return EU    (f, g, I, verbose); }
inline UBDD UBDD::EP    (const UBDD& f,                const bool& verbose) const { EY_Image I(*this); return EP    (f,    I, verbose); }
inline UBDD UBDD::AF    (const UBDD& f,                const bool& verbose) const { AX_Image I(*this); return AF    (f,    I, verbose); }
inline UBDD UBDD::AG    (const UBDD& f,                const bool& verbose) const { AX_Image I(*this); return AG    (f,    I, verbose); }
inline UBDD UBDD::AU    (const UBDD& f, const UBDD& g, const bool& verbose) const { AX_Image I(*this); return AU    (f, g, I, verbose); }
inline UBDD UBDD::EX_INF(const UBDD& f,                const bool& verbose) const { EX_Image I(*this); return EX_INF(f,    I, verbose); }
inline UBDD UBDD::EY_INF(const UBDD& f,                const bool& verbose) const { EY_Image I(*this); return EY_INF(f,    I, verbose); }
inline UBDD UBDD::FS    (const UBDD& f,                const bool& verbose) const { EY_Image I(*this); return FS    (f,    I, verbose); }
inline UBDD UBDD::BS    (const UBDD& f,                const bool& verbose) const { EX_Image I(*this); return BS    (f,    I, verbose); }

struct UBDD::Associative {
  const UBDD neutral; // returned when applying op to the empty set of elements
  ulong op_count;     // aux variable to keep track of number of op's applied
  culong total;       // total number of op's to be applied. For progress tracking
  inline Associative(const UBDD& neutral, culong& total = 0): neutral(neutral), op_count(0), total(total) {}
  virtual UBDD op(const UBDD& x, const UBDD& y) = 0;
  inline void update() {} // ++op_count; if (total > 0) printf("Associative::update: %ld op's (%.1f%%)\n", op_count, (100.0 * op_count) / total); }
  virtual ~Associative() {} };

inline UBDD UBDD::aggregate(const vector<UBDD>& factors, Associative& assoc) {
  return ( factors.empty() ? assoc.neutral : aggregate(factors, 0, factors.size(), assoc) ); }

// For convenience, some examples are provided
struct UBDD::Conjunction: public UBDD::Associative {
  inline Conjunction(culong& total = 0): Associative(UBDD::One(),  total) {}
  inline UBDD op(const UBDD& x, const UBDD& y) { update(); return x & y; } };

struct UBDD::Disjunction: public UBDD::Associative {
  inline Disjunction(culong& total = 0): Associative(UBDD::Zero(), total) {}
  inline UBDD op(const UBDD& x, const UBDD& y) { update(); return x | y; } };

inline UBDD UBDD::conjoin(const vector<UBDD>& conjuncts) { Conjunction c(conjuncts.size()); return aggregate(conjuncts, c); }
inline UBDD UBDD::disjoin(const vector<UBDD>& disjuncts) { Disjunction d(disjuncts.size()); return aggregate(disjuncts, d); }

///////////////////////////////////////////////////////////////////////////////
// Declaration of Tau classes (predicate transformers for fixpoint routines) //
///////////////////////////////////////////////////////////////////////////////

// tau with image operator
struct Tau_image: public Tau {
  Image& image;
  inline Tau_image(Image& image): image(image) {}
  virtual ~Tau_image() {} };

// tau suitable for EF, AF, EP. The predicate transformer needs to know f: tau(Z) = f | image(Z)
struct Tau_1: public Tau_image {
  const UBDD f;
  inline Tau_1(const UBDD& f, Image& image): Tau_image(image), f(f) {}
  inline UBDD operator()(const UBDD& Z) { return f | image(Z); } };

// tau suitable for EG, AG. The predicate transformer needs to know f: tau(Z) = f & image(Z)
struct Tau_2: public Tau_image {
  const UBDD f;
  inline Tau_2(const UBDD& f, Image& image): Tau_image(image), f(f) {}
  inline UBDD operator()(const UBDD& Z) { return f & image(Z); } };

// tau suitable for EU, AU. The predicate transformer needs to know f and g: tau(Z) = g | (f & image(Z))
struct Tau_3: public Tau_image {
  const UBDD f, g;
  inline Tau_3(const UBDD& f, const UBDD& g, Image& image): Tau_image(image), f(f), g(g) {}
  inline UBDD operator()(const UBDD& Z) { return g | (f & image(Z)); } };

// tau suitable for FS, BS. The predicate transformer needs to know nothing but the image function: tau(Z) = image(Z)
struct Tau_4: public Tau_image {
  inline Tau_4(Image& image): Tau_image(image) {}
  inline UBDD operator()(const UBDD& Z) { return image(Z); } };


//////////////////////////
// Non-member functions //
//////////////////////////

inline UBDD UBDD::EF    (const UBDD& f,                EX_Image& EX, const bool& verbose) { Tau_1 t(f,    EX); return Zero().FP    (t, "EF",     verbose); }
inline UBDD UBDD::EG    (const UBDD& f,                EX_Image& EX, const bool& verbose) { Tau_2 t(f,    EX); return One ().FP    (t, "EG",     verbose); }
inline UBDD UBDD::EU    (const UBDD& f, const UBDD& g, EX_Image& EX, const bool& verbose) { Tau_3 t(f, g, EX); return Zero().FP    (t, "EU",     verbose); }
inline UBDD UBDD::EP    (const UBDD& f,                EY_Image& EY, const bool& verbose) { Tau_1 t(f,    EY); return Zero().FP    (t, "EP",     verbose); }
inline UBDD UBDD::AF    (const UBDD& f,                AX_Image& AX, const bool& verbose) { Tau_1 t(f,    AX); return Zero().FP    (t, "AF",     verbose); }
inline UBDD UBDD::AG    (const UBDD& f,                AX_Image& AX, const bool& verbose) { Tau_2 t(f,    AX); return One ().FP    (t, "AG",     verbose); }
inline UBDD UBDD::AU    (const UBDD& f, const UBDD& g, AX_Image& AX, const bool& verbose) { Tau_3 t(f, g, AX); return Zero().FP    (t, "AU",     verbose); }
inline UBDD UBDD::EX_INF(const UBDD& f,                EX_Image& EX, const bool& verbose) { Tau_4 t(      EX); return   f   .FP    (t, "EX_INF", verbose); }
inline UBDD UBDD::EY_INF(const UBDD& f,                EY_Image& EY, const bool& verbose) { Tau_4 t(      EY); return   f   .FP    (t, "EY_INF", verbose); }
inline UBDD UBDD::FS    (const UBDD& f,                EY_Image& EY, const bool& verbose) { Tau_4 t(      EY); return   f   .Search(t, "FS",     verbose); }
inline UBDD UBDD::BS    (const UBDD& f,                EX_Image& EX, const bool& verbose) { Tau_4 t(      EX); return   f   .Search(t, "BS",     verbose); }

class Report {
  uint l;
  const string id; // name to be printed in front of output, for identification
  const bool verbose;
public:
  inline Report(const string& id, const bool& verbose = false): l(0), id(id), verbose(verbose) {}
  enum { NON_VERBOSE, VERBOSE };
  void print(const UBDD& iterate, ostream& out = cout); // non-const: updates l
  void print(const UBDD& Unexplored, const UBDD& Reached, ostream& out = cout); };

#endif
